#define	VGAPAINT_CAPTURE_C

#include <a/joy.h>

#include "vgadecod.h"
#include "capture.h"
#include "request.h"
#include "tools.h"
#include "io.h"
#include "listbox.h"
#include "picture.h"

/* The interrupt stub exists in ccycle.c */

void	CaptureDisplayRegs(void);

/* CaptureDisplay: Display results of a successful capture */

static char *RR_CRTC[]={
	"Horizontal Total",
	"Horizontal Disp End",
	"Start Horiz Blank",
	"End Horiz Blank",
	"Start Horiz Retrace",
	"End Horiz Retrace",
	"Vertical Total",
	"Overflow",
	"Preset Row Scan",
	"Max Scan Line",
	"Cursor Start",
	"Cursor End",
	"Start Addr High",
	"Start Addr Low",
	"Cursor Loc High",
	"Cursor Loc Low",
	"Vert Retrace Start",
	"Vert Retrace Low",
	"Vert Display End",
	"Offset",
	"Underline Loc",
	"Start Vert Blank",
	"End Vert Blank",
	"Mode Control",
	"Line Compare",
};

static char *RR_Seq[]={
	"Reset",
	"Clocking Mode",
	"Map Mask",
	"Char Map Select",
	"Memory Mode",
};

static char *RR_Graph[]={
	"Set/Reset",
	"Enable Set/Reset",
	"Color Compare",
	"Data Rotate",
	"Read Map Select",
	"Mode",
	"Miscellaneous",
	"Color Don't Care",
	"Bit Mask",
};

static char *RR_Attrib[]={
	"Palette 0",
	"Palette 1",
	"Palette 2",
	"Palette 3",
	"Palette 4",
	"Palette 5",
	"Palette 6",
	"Palette 7",
	"Palette 8",
	"Palette 9",
	"Palette A",
	"Palette B",
	"Palette C",
	"Palette D",
	"Palette E",
	"Palette F",
	"Mode Ctrl Reg",
	"Overscan Color",
	"Color Plane Enable",
	"Horiz Pixel Pan",
	"Color Bank Select",
};

static struct Zone Cd_Zn[]={
	{   2,187, 80, 11,ZN_3D+ZN_TEXTRENDER,NULL,"Delete Frame"	,NULL },
	{  91,187, 80, 11,ZN_3D+ZN_TEXTRENDER,NULL,"Registers"		,NULL },
	{ 191,187, 50, 11,ZN_3D+ZN_TEXTRENDER,NULL,"Decode"		,NULL },
	{ 268,187, 50, 11,ZN_3D+ZN_TEXTRENDER,NULL,"Extract"		,NULL },
	{ -1,-1,-1,-1 }
};

static struct ZoneRange Cd_Zr[]={
	{ NULL,Cd_Zn,NULL,-1,ZR_TAP,NULL,0,0,NULL,0 }
};

static struct Requester CdReq={
	0,0,320,200,
	REQ_BORDER+REQ_FORCECOL,
	"Screen capture results",
	St2,
	NULL,
	NULL,
	Cd_Zr,
	NULL,
	NULL,
	NULL
};

static char *CD_Pixel_Types[]={
	"16-color planar",
	"4-color chunky/planar",
	"256-color byte-per-pixel",
	"Unknown"
};

void	Capture_SimpleToAnim(void) {
	uchar *memptr;
	boolean first = TRUE;
	int pict=0;
	uchar *memptr0;

	// resize the picture to 320x200

	if (!PictureReinit(pict_cur,320,200)) return;	// uh oh
	picmem = pict_cur->frame_buf;
	piccol = pict_cur->palette;

	switch (vgacm.type) {
	case VCT_RFC:
		memptr = vgacm.static_mem;
		pict = vgacm.frames;
//pict = 1;
		while(pict--) {
			if (!first) if (!InsertAfterFrame(pict_cur)) return;
//Message("decompressing from %08lx",memptr);
			memptr0 = memptr;
			memptr = Frame_Decompress(pict_cur->frame_buf,memptr,64000,1,0);
			memptr += ((4-(memptr-memptr0))&3);
			if (!first) {
				if (memcmp(pict_cur->palette,memptr,768)) {
					if (!(pict_cur->current->frame_pal = malloc(768+32))) return;
					memcpy(pict_cur->current->frame_pal,memptr,768);
					memset(pict_cur->current->frame_pal+768,0xff,32);
				}
			} else {
				first=FALSE;
				memcpy(pict_cur->palette,memptr,768);
			}
			pict_cur->dirty = TRUE;
			memptr += 768;
			first = FALSE;
		}
		break;
	case VCT_RF:
		// work backwards from the last picture taken

		memptr = vgacm.static_ptr;
		pict = vgacm.frames;
		while(pict--) {
			memptr -= 64768;
			if (!first) {
				if (!InsertBeforeFrame(pict_cur)) return;
				if (memcmp(pict_cur->palette,memptr+64000,768)) {
//					if (!(pict_cur->current->frame_pal = malloc(768+32))) return;
					if (!AttachPaletteToFrame(pict_cur,pict_cur->current,NULL)) return;
					memcpy(pict_cur->current->frame_pal->palette,memptr+64000,768);
				}
			} else {
				first=FALSE;
				memcpy(pict_cur->palette,memptr+64000,768);
			}
			memcpy(pict_cur->frame_buf,memptr,64000);
			pict_cur->dirty = TRUE;

			// slice this picture off, but don't slice the last one off

//			if (pict)
//				realloc(vgacm.static_ptr,memptr-vgacm.static_ptr);
		}
		break;
	case VCT_RFX:
		// work backwards from the last picture taken

		memptr = vgacm.static_ptr;
		pict = vgacm.frames;
		while(pict--) {
			memptr -= 64768;
			if (!first) {
				if (!InsertBeforeFrame(pict_cur)) return;
				if (memcmp(pict_cur->palette,memptr+64000,768)) {
//					if (!(pict_cur->current->frame_pal = malloc(768+32))) return;
					if (!AttachPaletteToFrame(pict_cur,pict_cur->current,NULL)) return;
					memcpy(pict_cur->current->frame_pal->palette,memptr+64000,768);
				}
			} else {
				first=FALSE;
				memcpy(pict_cur->palette,memptr+64000,768);
			}
			{
				ulong x;
				ulong *dp = (ulong *)memptr;
				uchar *pp = pict_cur->frame_buf;
				int i;
				
				for(i=4000;i;i--) {
					x = dp[0];
					pp[    0] = (uchar)(x>>0);
					pp[    4] = (uchar)(x>>8);
					pp[    8] = (uchar)(x>>16);
					pp[   12] = (uchar)(x>>24);
					x = dp[4000];
					pp[    1] = (uchar)(x>>0);
					pp[    5] = (uchar)(x>>8);
					pp[    9] = (uchar)(x>>16);
					pp[   13] = (uchar)(x>>24);
					x = dp[8000];
					pp[    2] = (uchar)(x>>0);
					pp[    6] = (uchar)(x>>8);
					pp[   10] = (uchar)(x>>16);
					pp[   14] = (uchar)(x>>24);
					x = dp[12000];
					pp[    3] = (uchar)(x>>0);
					pp[    7] = (uchar)(x>>8);
					pp[   11] = (uchar)(x>>16);
					pp[   15] = (uchar)(x>>24);
					pp+=16;
					++dp;
				}
			}
			pict_cur->dirty = TRUE;

			// slice this picture off, but don't slice the last one off

//			if (pict)
//				realloc(vgacm.static_ptr,memptr-vgacm.static_ptr);
		}
		break;
	}
}

void	CaptureDisplay(void) {
	short pixmult;
	uchar ov;
	ushort start;
	uchar *dest=picmem,*src=capture->plane;
	long cnt;
	struct VGACaptureInfo vgaci;
	boolean quit=FALSE,decode;

	enum { CDPT_16P,CDPT_4CP,CDPT_256BPP } pix_type;

	if (!OpenRequester(&CdReq,0,0)) return;

	CaptureInterpret(capture,&vgaci);
	CaptureMutatePalette(capture->palette);

	pixmult=((capture->graph[5] & 0x60)==0x40) ? 4 : 8;
	ov=capture->crtc[7];
	start=capture->crtc[13]+((short)(capture->crtc[12])<<8);

	pix_type	= (capture->graph[5]>>5) & 3;

	vid_printf6(&CdReq.reqbm,4,14+0*8,St0,St2,"Scanned area: %d,%d pixels",
		capture->crtc[0]*pixmult,
		capture->crtc[6]+(ov & 1 ? 256 : 0)+(ov & 0x20 ? 512 : 0));
	vid_printf6(&CdReq.reqbm,4,14+1*8,St0,St2,"Displayed area: %d pixels, %d lines",
		(capture->crtc[1]+1)*pixmult,
		capture->crtc[18]+(ov & 2 ? 256 : 0)+(ov & 0x40 ? 512 : 0)+1);
	vid_printf6(&CdReq.reqbm,4,14+2*8,St0,St2,"Video area @ %04x, %ld bpr",
		start,capture->crtc[0x13]);
	vid_printf6(&CdReq.reqbm,4,14+3*8,St0,St2,"Pixel type: %s\n",
		(vgaci.flags & CDF_ALPHA) ? "alpha" : CD_Pixel_Types[pix_type]);
	vid_printf6(&CdReq.reqbm,4,14+4*8,St0,St2,"Line compare: %d\n",vgaci.plinecmp);

//	mouse_wait_key();
//	if (mouse_buttons() & 2) CaptureDisplayRegs();
//	mouse_wait_unkey();
	do {
		ulong msg;
		mouse_coord mx,my,mb;

		mouse_read2(&mx,&my,&mb);
		switch(msg=CheckRequester(&CdReq,mx,my,mb)) {
		case RZoneMsg(0,0):
			quit=TRUE;
			decode=FALSE;
			break;
		case RZoneMsg(0,1):
			CaptureDisplayRegs();
			break;
		case RZoneMsg(0,2):
			CaptureDecodeRegs();
			break;
		case RZoneMsg(0,3):
			quit=TRUE;
			decode=TRUE;
			break;
		}
	} while(!quit);

	CloseRequester(&CdReq);

	if (decode) {
		memcpy(piccol,capture->palette,768);
		if (capture->crtc[20] & 0x40) {		/* are we in mode X? */
			/* No. Unravel the bytes. */
			cnt=16000;
			do {
				*dest++=src[start];
				*dest++=src[start+0x10000];
				*dest++=src[start+0x20000];
				*dest++=src[start+0x30000];
				start+=4;
				if (!(start & 0x3FFC)) ++start;
			} while(--cnt);
		} else {
			/* Yes. Deinterleave the planes. */
			src+=start;
			cnt=16000;
			do {
				*dest++=src[0];
				*dest++=src[0x10000];
				*dest++=src[0x20000];
				*dest++=src[0x30000];
				++src;
			} while(--cnt);
		}
	}

	repick_colors();
	restore_screen();
}


//	CaptureMutatePalette: Convert colors from 6 bits to 8 bits

void	CaptureMutatePalette(uchar *p) {
	int i=768;

	do { *p=(*p*85L)/21; ++p;} while(--i);
}


void	CaptureInterpret(struct VGACapture *capt,struct VGACaptureInfo *cinf) {
	uchar	ov = capt->crtc[7];

	cinf->flags	= 0;

	if (!(capt->graph[6] & 1)) cinf->flags |= CDF_ALPHA;

	cinf->pix_type	= (capt->graph[5]>>5)&3;		// pixel type
	cinf->pix_mult	= (cinf->pix_type==CDPT_256BPP)?4:8;	// pixels per CRTC address
	cinf->pix_height= ((capt->crtc[9]&31)+1)*(capt->crtc[9]&0x80 ? 2 : 1);	// pixel height in scanlines
	cinf->pwidth	= (capt->crtc[1]+1)*cinf->pix_mult;	// pixel displayed width
	cinf->pheight	= (capt->crtc[18]+(ov & 2 ? 256 : 0)+(ov & 0x40 ? 512 : 0)+1);
	cinf->plinecmp	= capt->crtc[0x18] + ((ushort)(capt->crtc[9]&0x40)<<2) + ((ushort)(ov&0x10)<<5);
}


//////////////////////////// Capture Mode Requester ///////////////////////////

struct	Zone	CMReq_Z1[]={
	{ 56, 12,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_AUTOWIDTH				,NULL,"Simple",NULL },
	{  0, 12,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"RF",NULL },
	{  0, 12,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"RFX",NULL },
	{  0, 12,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"RFC",NULL },
	{  0, 12,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH+ZN_INVERTED	,NULL,"Advanced",NULL },
	{  0, 12,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"VESA",NULL },
	{ -1,-1,-1,-1 }
};

struct	Zone	CMReq_Z2[]={
	{ 56, 24,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_AUTOWIDTH+ZN_INVERTED		,NULL,"Int08",NULL },
	{  0, 24,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"Int09",NULL },
	{  0, 24,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"Int28",NULL },
	{ -1,-1,-1,-1 }
};

struct	Zone	CMReq_Z3[]={
	{ 56, 36,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_AUTOWIDTH				,NULL,"J1",NULL },
	{  0, 36,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"J2",NULL },
	{  0, 36,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"J3",NULL },
	{  0, 36,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"J4",NULL },
	{  0, 36,1, 11,ZN_TEXTRENDER+ZN_3D+ZN_FITRIGHT+ZN_AUTOWIDTH		,NULL,"PrtSc",NULL },
	{ -1,-1,-1,-1 }
};

struct Zone	CMReq_Z4[]={
	{   4, 74, 64, 24,ZN_3D+ZN_TEXTRENDER			,NULL,"Ok",NULL },
	{ 144, 74, 64, 24,ZN_3D+ZN_TEXTRENDER			,NULL,"Cancel",NULL },
	{ -1,-1,-1,-1 }
};

struct	ZoneRange CMReq_ZR[]={
	{ &CMReq_ZR[1]	,CMReq_Z1,&CMReq_Z1[3]	, 3,ZR_CHOICE,0,0,0,NULL,FALSE },
	{ &CMReq_ZR[2]	,CMReq_Z2,NULL		, 0,ZR_TOGGLE,0,0,0,NULL,FALSE },
	{ &CMReq_ZR[3]	,CMReq_Z3,NULL		, 0,ZR_TOGGLE,0,0,0,NULL,FALSE },
	{ NULL		,CMReq_Z4,NULL		, 0,ZR_TAP   ,0,0,0,NULL,FALSE },
};

struct	Text CMReq_Texts[]={
	{ "Type"	,  4, 14,St0,St2 },
	{ "Vectors"	,  4, 26,St0,St2 },
	{ "Triggers"	,  4, 38,St0,St2 },
	{ "Static"	,  4, 50,St0,St2 },
	{ "Dynamic"	,  4, 62,St0,St2 },

	{NULL}
};

void CMReq_SliderUpdater(struct Slider *sld) {
	uchar buffer[16];

	if (sld->cposx >= 99999)
		sprintf(buffer,"%4ldM",(sld->cposx+512)>>10);
	else
		sprintf(buffer,"%5ldK",sld->cposx);
	mouse_hide();
	vd->text6(sld->destbm,sld->arrow2.x + sld->arrow2.width+4,sld->boundbox.y,buffer,St0,St2);
	mouse_show();
}

struct	Slider CMReq_Sliders[]={
	{
		&CMReq_Sliders[1],{NULL},NULL,{ 70,50,86,8 },
		NULL,0,0,4,4,0,0,
		16384,0,16,0,0,0,0,0,
		SLD_OUTLINE3D+REQSLD_ONEHIT+SLD_ARROW1HIT+SLD_ARROW2HIT+SLD_ARROW1*4+SLD_ARROW2*6,
		&CMReq_SliderUpdater,
		{  56,48,12,12 },
		{ 158,48,12,12 },
		-1,0,1,0,
		NULL
	},{
		NULL,{NULL},NULL,{ 70,62,86,8 },
		NULL,0,0,4,4,0,0,
		16384,0,16,0,0,0,0,0,
		SLD_OUTLINE3D+REQSLD_ONEHIT+SLD_ARROW1HIT+SLD_ARROW2HIT+SLD_ARROW1*4+SLD_ARROW2*6,
		&CMReq_SliderUpdater,
		{  56,60,12,12 },
		{ 158,60,12,12 },
		-1,0,1,0,
		NULL
	}
};

static struct Requester CMReq={
	0,0,240,102,
	REQ_BORDER,
	"Capture Options",
	St2,
	NULL,
	NULL,
	CMReq_ZR,
	CMReq_Sliders,
	NULL,
	NULL,
	CMReq_Texts
};

boolean	CaptModeRequest(void) {
	mouse_coord x,y,b;
	ulong msg;
	boolean quit=FALSE,okay=FALSE;
	int	key;

	CMReq_Sliders[1].xclicks = 2 * (CMReq_Sliders[0].xclicks = (DPMI_MemFree()+1023)>>10);
	if (!OpenRequester(&CMReq,0,0)) return FALSE;

	do {
		key=kbd_check();
		if (key==' ' || key==0x1b) {
			okay = (key==' ');
			break;
		};

		mouse_read2(&x,&y,&b);
		if (b & 2) quit = TRUE;
		switch(msg=CheckRequester(&CMReq,x,y,b)) {
		case RZoneMsg(3,0):
			okay=TRUE;
		case RZoneMsg(3,1):
			quit=TRUE;
			break;
		};
	} while(!quit);
	CloseRequester(&CMReq);

	vgacm.joyMask = 0;
	if (CMReq_Z3[0].flags & ZN_INVERTED) vgacm.joyMask += JOY_A1;
	if (CMReq_Z3[1].flags & ZN_INVERTED) vgacm.joyMask += JOY_A2;
	if (CMReq_Z3[2].flags & ZN_INVERTED) vgacm.joyMask += JOY_B1;
	if (CMReq_Z3[3].flags & ZN_INVERTED) vgacm.joyMask += JOY_B2;

	vgacm.type		= CMReq_ZR[0].highindex;
	vgacm.max_static	= CMReq_Sliders[0].cposx * 1024;
	vgacm.max_dynamic	= CMReq_Sliders[1].cposx * 1024;

	return okay;
};

extern struct IHandler Capture_Int08;
extern struct IHandler Capture_Int09;

static char *shell_stack;
extern char far *shell_stack_ptr;

extern ushort shell_passup_segment;
extern ushort shell_routine_offset;
extern ushort shell_routine_segment;
extern ulong shell_routine[];

extern void shell_VPSTAT_Cmd(void);

void	Capture_Install(void) {
	union REGS regs;

	AddInterruptHandler(INT_TIMER,&Capture_Int08);
	vgacm.lastJoyRead = 0x00;

	// if we are using advanced capture, don't bother with static memory

	if (vgacm.type == VCT_ADVANCED)
		vgacm.static_ptr = vgacm.cur_static = 0;

	// attempt to allocate static memory - if we can't, report no static memory

	else if (vgacm.static_mem = malloc(vgacm.max_static)) {
		vgacm.static_ptr = vgacm.static_mem;
		vgacm.cur_static = 0;
	} else {
		vgacm.cur_static = vgacm.max_static;
	}

	// no pictures captured

	vgacm.frames = vgacm.framesbad = 0;
	vgacm.rapidfire = FALSE;
	vgacm.last_ptr = NULL;

	// install the DOS command VPSTAT

	if (shell_stack = malloc(10000)) {
		shell_stack_ptr = shell_stack+10000;
		if (shell_passup_segment = AllocCallback(&shell_VPSTAT_Cmd,FALSE)) {
			regs.w.ax = 0x0200;		// DPMI get real mode interrupt vector
			regs.h.bl = 0x2f;
			int386(0x31,&regs,&regs);

			shell_routine_segment = regs.w.cx;
			shell_routine_offset  = regs.w.dx;

			memcpy((char *)realmem+32,&shell_routine[1],shell_routine[0]);

			regs.w.ax = 0x0201;		// DPMI set real mode interrupt vector
			regs.h.bl = 0x2f;
			regs.w.cx = (ulong)((uchar *)realmem+32)>>4;
			regs.w.dx = (ulong)realmem&15;		// realmem is always para-aligned
			int386(0x31,&regs,&regs);
		} else free(shell_stack);
	}

#if 0
	{
		int i = shell_routine[0];
		char *p = (char *)&shell_routine[1];

		printf("shell_routine: %p\n", shell_routine);
		printf("shell_routine[0]: %08lx\n", shell_routine[0]);
		getch();
//		do {
//			printf("%02x ",*p++);
//		} while(--i);
	}
#endif
}

void	Capture_Deinstall(void) {
	union REGS regs;

	if (shell_passup_segment) {
		regs.w.ax = 0x0201;		// DPMI set real mode interrupt vector
		regs.h.bl = 0x2f;
		regs.w.cx = shell_routine_segment;
		regs.w.dx = shell_routine_offset;
		int386(0x31,&regs,&regs);

		free(shell_stack);
		FreeCallback(shell_passup_segment);
	}
	RemInterruptHandler(INT_TIMER,&Capture_Int08);
}

void	Capture_FreeMemory(void) {
	if (vgacm.static_mem) {
		free(vgacm.static_mem);
		vgacm.static_mem = NULL;
		_heapshrink();
	}
}

#pragma off (check_stack)

void	shell_VPSTAT_CmdC(void) {
	Capture_Print_Info();
}

/////////////////////////////////////////////////

static char *capt_str1[]={ "Simple","RapidFire","RapidFire-C","RapidFire-X","Advanced","VESA" };

void	Capture_Print_Info(void) {
	printf(	"Current capture mode:     %s\n"
		"Current capture triggers: joy(%c%c%c%c)\n"
		,capt_str1[vgacm.type]
		,(vgacm.joyMask & JOY_A1)?'1':' '
		,(vgacm.joyMask & JOY_A2)?'2':' '
		,(vgacm.joyMask & JOY_B1)?'3':' '
		,(vgacm.joyMask & JOY_B2)?'4':' '
		);
	if (vgacm.static_mem || !vgacm.max_static)
		printf(
			"Static memory:            %ld/%ld\n"
			,vgacm.cur_static,vgacm.max_static
			);
	else
		printf(	"Static memory:            *unable to allocate*\n");
	printf(	"Dynamic memory:           %ld/%ld\n"
		"Images captured/missed:   %ld/%ld\n"
		,vgacm.cur_dynamic,vgacm.max_dynamic
		,vgacm.frames,vgacm.framesbad
		);
}

void	CaptureDisplayRegs(void) {
	struct ListRequester *lr;
	int i;

	if (!(lr=MakeListRequester(40,18,"Capture: VGA Register Contents"))) return;

	lr->zones[0].text = "Okay";
	lr->zones[1].text = "Dump!";

	TextboxPrintf(lr->tbox,"VGA CRTC Registers");
	TextboxPrintf(lr->tbox,"------------------");
	for(i=0;i<0x19;i++)	TextboxPrintf(lr->tbox,"%-19s: %3d (%02x)",RR_CRTC[i],capture->crtc[i],capture->crtc[i]);
	TextboxPrintf(lr->tbox,"");
	TextboxPrintf(lr->tbox,"VGA Sequencer Registers");
	TextboxPrintf(lr->tbox,"-----------------------");
	for(i=0;i<0x05;i++)	TextboxPrintf(lr->tbox,"%-19s: %3d (%02x)",RR_Seq[i],capture->seq[i],capture->seq[i]);
	TextboxPrintf(lr->tbox,"");
	TextboxPrintf(lr->tbox,"VGA Graphics Controller Registers");
	TextboxPrintf(lr->tbox,"---------------------------------");
	for(i=0;i<0x09;i++)	TextboxPrintf(lr->tbox,"%-19s: %3d (%02x)",RR_Graph[i],capture->graph[i],capture->graph[i]);
	TextboxPrintf(lr->tbox,"");
	TextboxPrintf(lr->tbox,"VGA Attribute Controller Registers");
	TextboxPrintf(lr->tbox,"----------------------------------");
	for(i=0;i<0x15;i++)	TextboxPrintf(lr->tbox,"%-19s: %3d (%02x)",RR_Attrib[i],capture->attr[i],capture->attr[i]);
	TextboxPrintf(lr->tbox,"");
	TextboxPrintf(lr->tbox,"Other Registers");
	TextboxPrintf(lr->tbox,"---------------");
	TextboxPrintf(lr->tbox,"Miscellaneous Output: %3d (%02x)",capture->miscout,capture->miscout);
	TextboxPrintf(lr->tbox,"DAC Palette Mask:     %3d (%02x)",capture->palmask,capture->palmask);

	if (TextboxGlue(lr->tbox)) {
		TextboxCreate(lr->tbox,TRUE);
		mouse_wait_unkey();
		if (OpenRequester(lr->req,0,0)) {

			mouse_coord	mx,my,mb;
			boolean		cont = TRUE;
			ulong		msg;
			FILE		*f;
			char		**tp,*s;

			do {
				mouse_read2(&mx,&my,&mb);
				if (mb) {
					msg=CheckRequester(lr->req,mx,my,mb);
					mouse_wait_unkey();
					switch(msg) {
					case RZoneMsg(0,1):	// dump2disk

						if (f=VPfopen("VPRgDump.txt","wb")) {
							int x;

							tp=lr->tbox->table+1;

							for(x=0;x<lr->lbox[0].alen;x++) {
								fputs(*tp++,f);
								fputc('\r',f);
								fputc('\n',f);
							}

							fclose(f);
						};

					case RZoneMsg(0,0):	// ok
						cont=FALSE;
						break;
					}
				}
			} while(cont);

			CloseRequester(lr->req);
		}
		TextboxDissolve(lr->tbox);
	}
	DisposeListRequester(lr);
}

/////////////////////////////////////////////////////////

static char *vga_memory_modes[]={
	"odd/even","planar","chained","mode X"
};

static char *vga_autosize_modes[]={
	"reserved","400 lines","350 lines","480 lines"
};

static char *vga_clock_modes[]={
	"25 MHz","28 MHz","External 1","External 2"
};

static char *vga_write_modes[]={
	"0 (set/reset)","1 (vid2vid)","2 (fill)","3 (mask)"
};

void	CaptureDecodeRegs(void) {
	struct ListRequester *lr;
	struct VGADecode vgad;
	int i;

	Decode_VGA(capture,&vgad);

	if (!(lr=MakeListRequester(40,18,"Capture: VGA Decode Results"))) return;

	lr->zones[0].text = "Okay";
	lr->zones[1].text = "Dump!";

	TextboxPrintf(lr->tbox,"Absolute Screen Size:    %dx%d",vgad.horiz_total*vgad.pix_mult,vgad.vert_total);
	TextboxPrintf(lr->tbox,"Maximum Screen Size:     %dx%d",vgad.horiz_retrace_start*vgad.pix_mult,vgad.vert_retrace_start);
	TextboxPrintf(lr->tbox,"Overscan Screen Size:    %dx%d",vgad.horiz_blank_start*vgad.pix_mult,vgad.vert_blank_start);
	TextboxPrintf(lr->tbox,"Displayed Screen Size:   %dx%d",vgad.horiz_displayed*vgad.pix_mult,vgad.vert_displayed);
	TextboxPrintf(lr->tbox,"Logical width:           %d",vgad.logical_width);
	TextboxPrintf(lr->tbox,"Line compare:            %d",vgad.line_compare);
	TextboxPrintf(lr->tbox,"");
	TextboxPrintf(lr->tbox,"                           x     y");
	TextboxPrintf(lr->tbox,"                         ----- -----");
	TextboxPrintf(lr->tbox,"Blank interval start     %5d %5d",vgad.horiz_blank_start,vgad.vert_blank_start);
	TextboxPrintf(lr->tbox,"Blank interval end       %5d %5d",vgad.horiz_blank_end,vgad.vert_blank_end);
	TextboxPrintf(lr->tbox,"Retrace interval start   %5d %5d",vgad.horiz_retrace_start,vgad.vert_retrace_start);
	TextboxPrintf(lr->tbox,"Retrace interval end     %5d %5d",vgad.horiz_retrace_end,vgad.vert_retrace_end);
	TextboxPrintf(lr->tbox,"Monitor autosize         %s",vga_autosize_modes[(vgad.display_mode>>2)&3]);
	TextboxPrintf(lr->tbox,"Master clock             %s",vga_clock_modes[vgad.display_mode&3]);
	TextboxPrintf(lr->tbox,"");
	TextboxPrintf(lr->tbox,"Number of rows:          %d",vgad.row_count);
	TextboxPrintf(lr->tbox,"Vertical fine scroll:    %d",vgad.row_fine_scroll);
	TextboxPrintf(lr->tbox,"Underline location:      %d",vgad.row_underline);
	TextboxPrintf(lr->tbox,"Cursor start/end:        %d-%d",vgad.cursor_start,vgad.cursor_end);
	TextboxPrintf(lr->tbox,"");
	TextboxPrintf(lr->tbox,"Start address:           %04x",vgad.start_address);
	TextboxPrintf(lr->tbox,"Cursor address:          %04x",vgad.cursor_address);
	TextboxPrintf(lr->tbox,"Aperture start:          %05lx-%05lx",vgad.aperture_start,(char *)vgad.aperture_start+vgad.aperture_size-1);
	TextboxPrintf(lr->tbox,"Aperture size:           %ldK",vgad.aperture_size>>10);
	TextboxPrintf(lr->tbox,"I/O aperture start:      %03x",vgad.aperture_io);
	TextboxPrintf(lr->tbox,"Memory mode:             %s",vga_memory_modes[vgad.memory_mode]);
	TextboxPrintf(lr->tbox,"");
	TextboxPrintf(lr->tbox,"Graphics write mode:     %s",vga_write_modes[vgad.write_mode]);

	if (TextboxGlue(lr->tbox)) {
		TextboxCreate(lr->tbox,TRUE);
		mouse_wait_unkey();
		if (OpenRequester(lr->req,0,0)) {

			mouse_coord	mx,my,mb;
			boolean		cont = TRUE;
			ulong		msg;
			FILE		*f;
			char		**tp,*s;

			do {
				mouse_read2(&mx,&my,&mb);
				if (mb) {
					msg=CheckRequester(lr->req,mx,my,mb);
					mouse_wait_unkey();
					switch(msg) {
					case RZoneMsg(0,1):	// dump2disk

						if (f=VPfopen("VPDcDump.txt","wb")) {
							int x;

							tp=lr->tbox->table+1;

							for(x=0;x<lr->lbox[0].alen;x++) {
								fputs(*tp++,f);
								fputc('\r',f);
								fputc('\n',f);
							}

							fclose(f);
						};

					case RZoneMsg(0,0):	// ok
						cont=FALSE;
						break;
					}
				}
			} while(cont);

			CloseRequester(lr->req);
		}
		TextboxDissolve(lr->tbox);
	}
	DisposeListRequester(lr);
}



