#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <dos.h>
#include <math.h>

#include <a/types.h>

#include "tokenize.h"

#define M_PI 3.1415926535897932384626433832

double angle(double,double);

short math_error;

double	equ_eval_dbl(short *tokstream,double *stack,double *vars) {
	short tok;
	short *tk=tokstream;

	math_error=0;

	while(tok=*tokstream++) {
		switch(tok) {
		case EVAL_VAL:
			*stack++=*(double*)tokstream;
			tokstream=(short*)((char*)tokstream+sizeof(double));
			break;
		case EVAL_VAR:	*stack++=vars[*tokstream++]; break;
		case EVAL_NEG:	stack[-1]=-stack[-1]; break;
		case EVAL_ADD:	stack[-2]+=stack[-1]; --stack; break;
		case EVAL_SUB:	stack[-2]-=stack[-1]; --stack; break;
		case EVAL_MUL:	stack[-2]*=stack[-1]; --stack; break;
		case EVAL_DIV:	stack[-2]/=stack[-1]; --stack; break;
		case EVAL_POW:	stack[-2]=pow(stack[-2],stack[-1]); --stack; break;
		case EVAL_SIN:	stack[-1]=sin(stack[-1]); break;
		case EVAL_COS:	stack[-1]=cos(stack[-1]); break;
		case EVAL_TAN:	stack[-1]=tan(stack[-1]); break;
		case EVAL_SEC:	stack[-1]=1/cos(stack[-1]); break;
		case EVAL_CSC:	stack[-1]=1/sin(stack[-1]); break;
		case EVAL_ASIN:	stack[-1]=asin(stack[-1]); break;
		case EVAL_ACOS:	stack[-1]=acos(stack[-1]); break;
		case EVAL_ATAN:	stack[-1]=atan(stack[-1]); break;
		case EVAL_ASEC:	stack[-1]=acos(1/stack[-1]); break;
		case EVAL_ACSC: stack[-1]=asin(1/stack[-1]); break;
		case EVAL_ACOT:	stack[-1]=atan(1/stack[-1]); break;
		case EVAL_LOG:	stack[-1]=log(stack[-1]); break;
		case EVAL_CLOG:	stack[-1]=log10(stack[-1]); break;
		case EVAL_EXP:	stack[-1]=pow(2.718281828459045,stack[-1]); break;
		case EVAL_SQRT:	stack[-1]=sqrt(stack[-1]); break;
		case EVAL_PI:	*stack++=M_PI; break;
		case EVAL_POL:	vars['I'-'A']=sqrt(stack[-2]*stack[-2]+stack[-1]*stack[-1]);
				vars['J'-'A']=angle(stack[-2],stack[-1]);
				stack[-2]=vars['I'-'A'];
				--stack;
				break;
		case EVAL_REC:	vars['I'-'A']=stack[-2]*cos(stack[-1]);
				vars['J'-'A']=stack[-2]*sin(stack[-1]);
				stack[-2]=vars['I'-'A'];
				--stack;
				break;
		case EVAL_ASG:	vars[*tokstream++]=stack[-1]; break;
		case EVAL_SYNEVAL:	{
					short parmcnt=*tokstream++;
					double val,accum=0.0;
					double *sptr;

					sptr=&stack[-parmcnt];
					val=sptr[-1];

					do {
						accum*=val;
						accum+=*sptr++;
					} while(--parmcnt);

					stack-=parmcnt;
					stack[-1]=accum;
				}
				break;
		case EVAL_ABS:	stack[-1]=fabs(stack[-1]); break;
		case EVAL_INT:	modf(stack[-1],stack-1); break;
		case EVAL_POP:	--stack; break;
		case EVAL_EQUAL:	stack[-2]=(stack[-2]==stack[-1]); --stack; break;
		case EVAL_LESS:		stack[-2]=(stack[-2]< stack[-1]); --stack; break;
		case EVAL_GRTR:		stack[-2]=(stack[-2]> stack[-1]); --stack; break;
		case EVAL_LESSEQ:	stack[-2]=(stack[-2]<=stack[-1]); --stack; break;
		case EVAL_GRTREQ:	stack[-2]=(stack[-2]>=stack[-1]); --stack; break;
		};

		if (math_error) break;
	}

	return stack[-1];
}

double angle(double x,double y) {
	double a1;

	if (x==0) if (y>0) return M_PI/2; else if (y<0) return (M_PI*3)/2; else return 0;

	if (abs(y)>=abs(x)) {
		a1=atan(y/abs(x));
	} else
		a1=(y<0 ? ((M_PI*3)/2) : (M_PI/2))-atan((double)abs(x)/(double)y);

	if (x<0) a1=M_PI-a1;
	if (a1<0) a1+=2*M_PI;

	return a1;
}

int matherr(struct exception *err_info) {
	switch(err_info->type) {
	case DOMAIN:
	case SING:
	case OVERFLOW:
		math_error=err_info->type;
	case UNDERFLOW:
	case TLOSS:
	case PLOSS:
		break;
	};
	return 1;
}

