%{
	#include <stdio.h>
	#include <stdlib.h>
	#include <stddef.h>
	#include <math.h>

	#include <a/types.h>

	#include "tokenize.h"

	extern short *equ_output;
%}

%token	FLEXERR

%token	SINE		COSINE		TANGENT		SECANT
%token	COSECANT	COTANGENT	ARC_SINE	ARC_COSINE
%token	ARC_TANGENT	ARC_SECANT	ARC_COSECANT	ARC_COTANGENT
%token	POL_TO_REC	REC_TO_POL	LOG		CLOG
%token	EXP		SQRT

%token	PI	SYNEVAL

%token	ABS	INT

%token	EQUAL	LESS	GRTR	LESSEQ	GRTREQ

%token	<var>	VAR
%token	<val>	VALUE

%union {
	char	var;
	short	count;
	double	val;
}

%left	':'
%right	'='
%left	EQUAL LESS GRTR LESSEQ GRTREQ
%left	'+' '-'
%left	'*' '/'
%left	SINE COSINE TANGENT SECANT COSECANT COTANGENT ARC_SINE ARC_COSINE ARC_TANGENT
	ARC_SECANT ARC_COSECANT ARC_COTANGENT LOG CLOG EXP SQRT
%right	'^'
%right	NEG

%type	<count>	synparm

%%

expr:	  iexpr2
	| '-' iexpr2 %prec NEG	{ evat(EVAL_NEG) }
	| expr ':' { evat(EVAL_POP) } expr
	| expr '+' expr	{ evat(EVAL_ADD) }
	| expr '-' expr { evat(EVAL_SUB) }
	| expr '*' expr { evat(EVAL_MUL) }
	| expr '/' expr { evat(EVAL_DIV) }
	| expr '^' expr { evat(EVAL_POW) }
	| expr EQUAL expr	{ evat(EVAL_EQUAL) }
	| expr LESS expr	{ evat(EVAL_LESS) }
	| expr GRTR expr	{ evat(EVAL_GRTR) }
	| expr LESSEQ expr	{ evat(EVAL_LESSEQ) }
	| expr GRTREQ expr	{ evat(EVAL_GRTREQ) }
	| VAR '=' expr	{ evat(EVAL_ASG) evat($1) }
	;

iexpr2:	  atoms
	| iexprx
	| iexpr2 iexprx		{ evat(EVAL_MUL); }
	;

iexpr:	  iexpr1
	| '-' iexpr1		{ evat(EVAL_NEG) }

iexpr1:	  atoms
	| iexprx
	;

iexprx:	  SINE iexpr		{ evat(EVAL_SIN);  		}
	| COSINE iexpr		{ evat(EVAL_COS);  		}
	| TANGENT iexpr		{ evat(EVAL_TAN);  		}
	| SECANT iexpr		{ evat(EVAL_SEC);  		}
	| COSECANT iexpr	{ evat(EVAL_CSC);  		}
	| COTANGENT iexpr	{ evat(EVAL_COT);		}
	| ARC_SINE iexpr	{ evat(EVAL_ASIN); 		}
	| ARC_COSINE iexpr	{ evat(EVAL_ACOS);		}
	| ARC_TANGENT iexpr	{ evat(EVAL_ATAN);		}
	| ARC_SECANT iexpr	{ evat(EVAL_ASEC);		}
	| ARC_COSECANT iexpr	{ evat(EVAL_ACSC);		}
	| ARC_COTANGENT iexpr	{ evat(EVAL_ACOT);		}
	| LOG iexpr		{ evat(EVAL_LOG);  		}
	| CLOG iexpr		{ evat(EVAL_CLOG); 		}
	| EXP iexpr		{ evat(EVAL_EXP);  		}
	| SQRT iexpr		{ evat(EVAL_SQRT); 		}
	| INT iexpr		{ evat(EVAL_INT);		}
	| ABS iexpr		{ evat(EVAL_ABS);		}
	| SYNEVAL '(' expr ',' synparm ')'	{ evat(EVAL_SYNEVAL); evat($5) }
	| POL_TO_REC '(' expr ',' expr ')'	{ evat(EVAL_REC); }
	| REC_TO_POL '(' expr ',' expr ')'	{ evat(EVAL_POL); }
	;

atoms:	  atoms atom	{ evat(EVAL_MUL) }
	| atom
	;

atom:	  VAR		{ evat(EVAL_VAR); evat($1); }
	| PI		{ evat(EVAL_PI); }
	| '(' expr ')'
	| VALUE		{ evat(EVAL_VAL) evad($1) }
	;

synparm:	  expr			{ $$=1;		}
		| synparm ',' expr	{ $$=$1+1;	}

%%

