#include <malloc.h>

#include <a/bits.h>

#include "vgapaint.h"
#include "info.h"
#include "io.h"
#include "listbox.h"
#include "picture.h"
#include "anim.h"

struct meminfo {
    unsigned LargestBlockAvail;
    unsigned MaxUnlockedPage;
    unsigned LargestLockablePage;
    unsigned LinAddrSpace;
    unsigned NumFreePagesAvail;
    unsigned NumPhysicalPagesFree;
    unsigned TotalPhysicalPages;
    unsigned FreeLinAddrSpace;
    unsigned SizeOfPageFile;
    unsigned Reserved[3];
} MemInfo;

static struct BenchMark {
	ulong xms_bpsr;		/* >1Mb memory bytes/second */
	ulong xms_bpsw;		/* >1Mb memory bytes/second */
	ulong dos_bpsr;		/* <1Mb memory bytes/second */
	ulong dos_bpsw;		/* <1Mb memory bytes/second */
	ulong vid_bpsr;		/* video bytes/second */
	ulong vid_bpsw;		/* video bytes/second */
	ulong io_bpsr;		/* I/O bytes/second (DAC color register) */
	ulong io_bpsw;		/* I/O bytes/second (DAC color register) */
	ulong vidx_bpsr;	/* video bytes/second (mode X) */
	ulong vidx_bpsw;	/* video bytes/second (mode X) */
	ulong vidb_bpsr;	/* video bytes/second (blanked) */
	ulong vidb_bpsw;	/* video bytes/second (blanked) */
	ulong viddrv_write;
} benchm;

static struct Rectangle InfoRects[]={
	{  42,162, 64, 11 },
	{ 214,162, 64, 11 },
	{ RECT_END,RECT_END,RECT_END,RECT_END },
};

ulong stackptr(void);
#pragma aux stackptr="mov eax,esp" modify exact [eax] nomemory value [eax];

void Benchmark(char **);

void Info(void) {
	mouse_coord x,y,b;
	int rect;
	ulong	t;

	vid_loadRGB6(0,stdcol,32);
	col=col_std;
	mouse_hide();
	vd->bar(&vidbm,40,25,240,150,ST2);
	vid_draw3d(&vidbm,40,25,279,174);
	vd->text6(&vidbm,160-3*17,27,"VGAPaint 386 Info",ST0,ST2);
	vd->hline(&vidbm,42,277,37,ST1);
	vd->hline(&vidbm,42,277,38,ST3);
	vid_draw3d(&vidbm,42,162,105,172);
	vd->text6(&vidbm,42+32-9*3,164,"Benchmark",ST0,ST2);
	vid_draw3d(&vidbm,214,162,277,172);
	vd->text6(&vidbm,214+32-4*3,164,"Okay",ST0,ST2);

	{
		char *str;
		char cpud;

		switch(cpud=cpu_detect()) {
		case 3:	str="80386"; break;
		case 4:	str="80486"; break;
		case 5: str="Pentium";	break;
		default:	str="Unknown"; break;
		};
		vid_printf6(&vidbm, 44,42+  0,ST0,ST2,"CPU Type:    %s-%d",str,(cpu_speed(cpud)+500000)/1000000);
	}
	vid_printf6(&vidbm, 44,42+  8,ST0,ST2,"Free memory: %ld bytes",MemFree());
	{
		ulong msg;

		vid_printf6(&vidbm, 44,42+ 16,ST0,ST2,"OS:");

		if (sys.LinuxDOSEMU) {
			vid_printf6(&vidbm,122,42+ 16,ST0,ST2,"Linux DOSEMU");
		} else {
			msg=DetectMWindows();

			if (msg == 20)
				vid_printf6(&vidbm,122,42+ 16,ST0,ST2,"OS/2");
			else if (msg==0x0200)
				vid_printf6(&vidbm,122,42+ 16,ST0,ST2,"Windows/386 %d.xx",msg>>8);
			else if (msg>0x0200)
				vid_printf6(&vidbm,122,42+ 16,ST0,ST2,"Windows %d.%d",msg>>8,msg&255);
			else
				vid_printf6(&vidbm,122,42+ 16,ST0,ST2,"DOS");
		}
	}
	vid_printf6(&vidbm, 44,42+ 24,ST0,ST2,"VESA support:  %s",sys.VESA ? (sys.VESA2_0 ? "VESA 2.0" : "VESA 1.x") : "No");
	vid_printf6(&vidbm, 44,42+ 32,ST0,ST2,"Picture:       %ldx%ld (%ldK)",picbm->width,picbm->height,(picbm->size+1023)>>10);

	t=AnimSize(pict_cur);
	vid_printf6(&vidbm, 44,42+ 40,ST0,ST2,"  %ld frames, %ld bytes (%ldK)",pict_cur->total_frames,t,(t+1023)>>10);
	mouse_show();

	for(;;) {
		mouse_wait_key();
		mouse_read2(&x,&y,&b);
		switch(rect=rectSearch(&InfoRects,x,y)) {
		case 0:
//			Benchmark();

			{
				struct ListRequester *lr;

				if (lr=MakeListRequester(40,8,"Benchmark results")) {
					TextboxPrintf(lr->tbox,"XMS mem read bytes/sec:   < disabled >");
					TextboxPrintf(lr->tbox,"XMS mem write bytes/sec:  < disabled >");
					TextboxPrintf(lr->tbox,"DOS mem read bytes/sec:   < disabled >");
					TextboxPrintf(lr->tbox,"DOS mem write bytes/sec:  < disabled >");
					TextboxPrintf(lr->tbox,"Video read bytes/sec:     < disabled >");
					TextboxPrintf(lr->tbox,"Video write bytes/sec:    < disabled >");
					TextboxPrintf(lr->tbox,"Video DAC read bytes/sec: < disabled >");
					TextboxPrintf(lr->tbox,"Video DAC write bps:      < disabled >");
					TextboxPrintf(lr->tbox,"Video read bps Mode X:    < disabled >");
					TextboxPrintf(lr->tbox,"Video write bps Mode X:   < disabled >");
					TextboxPrintf(lr->tbox,"Video read bps blanked:   < disabled >");
					TextboxPrintf(lr->tbox,"Video write bps blanked:  < disabled >");
					TextboxPrintf(lr->tbox,"Video driver write speed: < disabled >");
					if (TextboxGlue(lr->tbox)) {
						mouse_coord	mx,my,mb;
						ulong		msg;
						boolean		cont=TRUE;

						TextboxCreate(lr->tbox,TRUE);

						lr->zones[1].text	= "Bench!";

						mouse_wait_unkey();

						if (!OpenRequester(lr->req,0,0)) break;

						do {
							mouse_read2(&mx,&my,&mb);
							if (mb & 1) {
								msg=CheckRequester(lr->req,mx,my,mb);
								mouse_wait_unkey();
								switch(msg) {
								case RZoneMsg(0,0):	/* Ok */
									cont=FALSE;
									break;
								case RZoneMsg(0,1):	/* Cancel */
									Benchmark((char **)lr->lbox->listadd);
									{
										short x;
										ulong *pp=(ulong *)&benchm;
										char **s=(char **)lr->lbox->listadd;

										for(x=13;x;x--,++s,++pp) if ((*s)[strlen(*s)-10]!='d') sprintf(*s+strlen(*s)-12,"%12ld",*pp);

										mouse_hide();
										DrawListbox(lr->lbox);
										mouse_show();
									}
									break;
								case REQM_LISTBOX+0:
									if (lr->lbox->sel >= 0) {
										char *s;

										s=((char **)lr->lbox->listadd)[lr->lbox->sel];

										if (s) {	// enabled?
											ulong *t = (ulong *)(s+strlen(s)-10);

											if (*(uchar *)t=='d')
												t[0]='ane ';	// "ena "
											else
												t[0]='asid';	// "disa"
											t[1] = 'delb';
											*(ushort *)(t+2) = '> ';
											((ushort *)t)[-1] = ' <';

											mouse_hide();
											DrawListbox(lr->lbox);
											mouse_show();
										}
									}
									break;
								};
							};
						} while(cont);

						CloseRequester(lr->req);

						TextboxDissolve(lr->tbox);
					};
					DisposeListRequester(lr);
				};
			}

			break;
		case 1: goto info_exit;
		};
		mouse_wait_unkey();
	};
info_exit:
	restore_screen();
}

ulong MemFree(void) {
	union REGS regs;
	struct SREGS sregs;
	size_t memsiz;

	memsiz=_memavl();

	regs.x.eax = 0x00000500;
	memset( &sregs, 0, sizeof(sregs) );
	sregs.es = FP_SEG( &MemInfo );
	regs.x.edi = FP_OFF( &MemInfo );

	int386x(0x31, &regs, &regs, &sregs );

	return memsiz+(MemInfo.NumFreePagesAvail<<12);
}

ulong Benchmark_memoryrbps(ulong maxcount,ulong *countptr,uchar *memory);
#pragma	aux Benchmark_memoryrbps=	\
	"push ebp",		\
	"push eax",		\
	"mov ebp,eax",		\
	"xor ebp,ebp",		\
	"mov ebx,0046ch",	\
	"mov ebx,[ebx]",	\
	"l:",			\
	"inc ebp",		\
	"mov esi,edi",		\
	"mov ecx,00004000h",	\
	"rep lodsd",		\
	"mov eax,0046ch",	\
	"mov eax,[eax]",	\
	"sub eax,ebx",		\
	"cmp eax,edx",		\
	"jb  l",		\
	"pop ecx",		\
	"mov [ecx],ebp",	\
	"pop ebp",		\
	parm [edx] [eax] [edi]	\
	value [eax]		\
	modify [eax ebx ecx edx esi edi];

ulong Benchmark_memorywbps(ulong maxcount,ulong *countptr,uchar *memory);
#pragma	aux Benchmark_memorywbps=	\
	"push ebp",		\
	"push eax",		\
	"mov ebp,eax",		\
	"xor ebp,ebp",		\
	"mov ebx,0046ch",	\
	"mov ebx,[ebx]",	\
	"l:",			\
	"inc ebp",		\
	"mov edi,esi",		\
	"mov ecx,00004000h",	\
	"rep stosd",		\
	"mov eax,0046ch",	\
	"mov eax,[eax]",	\
	"sub eax,ebx",		\
	"cmp eax,edx",		\
	"jb  l",		\
	"pop ecx",		\
	"mov [ecx],ebp",	\
	"pop ebp",		\
	parm [edx] [eax] [esi]	\
	value [eax]		\
	modify [eax ebx ecx edx esi edi];

ulong Benchmark_vidrbps(ulong maxcount,ulong *countptr);
#pragma	aux Benchmark_vidrbps=	\
	"push ebp",		\
	"push eax",		\
	"mov edi,000a0000h",	\
	"mov ebp,eax",		\
	"xor ebp,ebp",		\
	"mov ebx,0046ch",	\
	"mov ebx,[ebx]",	\
	"l:",			\
	"inc ebp",		\
	"mov esi,edi",		\
	"mov ecx,00004000h",	\
	"rep lodsd",		\
	"mov eax,0046ch",	\
	"mov eax,[eax]",	\
	"sub eax,ebx",		\
	"cmp eax,edx",		\
	"jb  l",		\
	"pop ecx",		\
	"mov [ecx],ebp",	\
	"pop ebp",		\
	parm [edx] [eax]	\
	value [eax]		\
	modify [eax ebx ecx edx esi edi];

ulong Benchmark_vidwbps(ulong maxcount,ulong *countptr);
#pragma	aux Benchmark_vidwbps=	\
	"push ebp",		\
	"push eax",		\
	"mov esi,000a0000h",	\
	"mov ebp,eax",		\
	"xor ebp,ebp",		\
	"mov ebx,0046ch",	\
	"mov ebx,[ebx]",	\
	"l:",			\
	"inc ebp",		\
	"mov edi,esi",		\
	"mov ecx,00004000h",	\
	"rep stosd",		\
	"mov eax,0046ch",	\
	"mov eax,[eax]",	\
	"sub eax,ebx",		\
	"cmp eax,edx",		\
	"jb  l",		\
	"pop ecx",		\
	"mov [ecx],ebp",	\
	"pop ebp",		\
	parm [edx] [eax]	\
	value [eax]		\
	modify [eax ebx ecx edx esi edi];

ulong Benchmark_vidiorbps(ulong maxcount,ulong *countptr);
#pragma	aux Benchmark_vidiorbps=	\
	"push ebp",		\
	"push eax",		\
	"mov ebp,eax",		\
	"xor ebp,ebp",		\
	"mov ebx,0046ch",	\
	"mov ebx,[ebx]",	\
	"mov edx,000003c9h",	\
	"l:",			\
	"inc ebp",		\
	"mov edi,offset packmem",	\
	"mov ecx,00000300h",	\
	"rep insb",		\
	"mov eax,0046ch",	\
	"mov eax,[eax]",	\
	"sub eax,ebx",		\
	"cmp eax,esi",		\
	"jb  l",		\
	"pop ecx",		\
	"mov [ecx],ebp",	\
	"pop ebp",		\
	parm [esi] [eax]	\
	value [eax]		\
	modify [eax ebx ecx edx esi edi];

ulong Benchmark_vidiowbps(ulong maxcount,ulong *countptr);
#pragma	aux Benchmark_vidiowbps=	\
	"push ebp",		\
	"push eax",		\
	"mov ebp,eax",		\
	"xor ebp,ebp",		\
	"mov ebx,0046ch",	\
	"mov ebx,[ebx]",	\
	"mov edx,000003c9h",	\
	"l:",			\
	"inc ebp",		\
	"mov esi,offset packmem",	\
	"mov ecx,00000300h",	\
	"rep outsb",		\
	"mov eax,0046ch",	\
	"mov eax,[eax]",	\
	"sub eax,ebx",		\
	"cmp eax,edi",		\
	"jb  l",		\
	"pop ecx",		\
	"mov [ecx],ebp",	\
	"pop ebp",		\
	parm [edi] [eax]	\
	value [eax]		\
	modify [eax ebx ecx edx esi edi];


boolean is_enabled(char *s) {
	return s[strlen(s)-10]!='d';
}

void Benchmark(char **pp) {
	BLOCK bl;
	ulong cnt,cnt2;

	uchar *mem;
	ushort realdsc;

	mouse_hide();
	bl=vd->blocksave(&vidbm,0,0,320,200);

	{
		coord x,y,xs,ys;

		xs = sysfontwidth*15;
		ys = sysfontheight;
		x = (vidbm.width -xs)>>1;
		y = (vidbm.height-ys)>>1;

		mouse_hide();
		vd->bar(&vidbm,x-2*XPOINT,y-2*YPOINT,xs+4*XPOINT,ys+4*YPOINT,St2);
		vid_draw3d(&vidbm,x-2*XPOINT,y-2*YPOINT,x+xs+2*XPOINT-1,y+ys+2*YPOINT-1);
		SysText(&vidbm,x,y,"Benchmarking...",St0,St2);
		mouse_show();
	}

	if (mem=malloc(65536)) {
		if (is_enabled(*pp++)) {
			cnt2=Benchmark_memoryrbps(91,&cnt,mem);
			benchm.xms_bpsr=(65536.0*(1193181.0/65536.0)*cnt)/cnt2;
		};

		if (is_enabled(*pp++)) {
			cnt2=Benchmark_memorywbps(91,&cnt,mem);
			benchm.xms_bpsw=(65536.0*(1193181.0/65536.0)*cnt)/cnt2;
		};

		free(mem);
	};

	if (mem=DosMemoryAllocate(4096,&realdsc)) {

		if (is_enabled(*pp++)) {
			cnt2=Benchmark_memoryrbps(91,&cnt,mem);
			benchm.dos_bpsr=(65536.0*(1193181.0/65536.0)*cnt)/cnt2;
		};
		if (is_enabled(*pp++)) {
			cnt2=Benchmark_memorywbps(91,&cnt,mem);
			benchm.dos_bpsw=(65536.0*(1193181.0/65536.0)*cnt)/cnt2;
		};

		DosMemoryFree(realdsc);
	};

	if (is_enabled(*pp++)) {
		cnt2=Benchmark_vidrbps(91,&cnt);
		benchm.vid_bpsr=(65536.0*(1193181.0/65536.0)*cnt)/cnt2;
	};

	if (is_enabled(*pp++)) {
		cnt2=Benchmark_vidwbps(91,&cnt);
		benchm.vid_bpsw=(65536.0*(1193181.0/65536.0)*cnt)/cnt2;
	};

	memset(packmem,0xFF,768);

	if (is_enabled(*pp++)) {
		cnt2=Benchmark_vidiorbps(91,&cnt);
		benchm.io_bpsr=(768.0*(1193181.0/65536.0)*cnt)/cnt2;
	};

	memcpy(packmem,piccol,768);

	if (is_enabled(*pp++)) {
		outportb(0x3c8,0);
		cnt2=Benchmark_vidiowbps(91,&cnt);
		benchm.io_bpsw=(768.0*(1193181.0/65536.0)*cnt)/cnt2;
	};

	vid_loadRGB6(0,piccol,256);

	/* jump to Mode X */

	disable();
	outportb(0x3d4,0x14);
	outportb(0x3d5,0);
	outportb(0x3d4,0x17);
	outportb(0x3d5,0xe3);
	outportb(0x3c4,4);
	outportb(0x3c5,6);
	enable();

	if (is_enabled(*pp++)) {
		cnt2=Benchmark_vidrbps(91,&cnt);
		benchm.vidx_bpsr=(65536.0*(1193181.0/65536.0)*cnt)/cnt2;
	};

	if (is_enabled(*pp++)) {
		cnt2=Benchmark_vidwbps(91,&cnt);
		benchm.vidx_bpsw=(65536.0*(1193181.0/65536.0)*cnt)/cnt2;
	};

	/* jump back to Mode 13h */

	disable();
	outportb(0x3d4,0x14);
	outportb(0x3d5,0x40);
	outportb(0x3d4,0x17);
	outportb(0x3d5,0xa3);
	outportb(0x3c4,4);
	outportb(0x3c5,0x0E);

	/* tell the sequencer to shut off access from CRTC to speed access */

//	outportb(0x3c4,0x01);
//	temp=inportb(0x3c5);
//	outportb(0x3c4,0x01);
//	outportb(0x3c5,temp | 0x20);
	enable();

	regs.w.ax=0x1201;
	regs.h.bl=0x36;
	int386(0x10,&regs,&regs);

	if (is_enabled(*pp++)) {
		cnt2=Benchmark_vidrbps(91,&cnt);
		benchm.vidb_bpsr=(65536.0*(1193181.0/65536.)*cnt)/cnt2;
	};

	if (is_enabled(*pp++)) {
		cnt2=Benchmark_vidwbps(91,&cnt);
		benchm.vidb_bpsw=(65536.0*(1193181.0/65536.)*cnt)/cnt2;
	};

	/* tell the sequencer to restore access to CRTC */

	regs.w.ax=0x1200;
	regs.h.bl=0x36;
	int386(0x10,&regs,&regs);

//	disable();
//	outportb(0x3c4,0x01);
//	outportb(0x3c5,temp);
//	enable();

	if (is_enabled(*pp++)) {
		int i;
		ushort t1,t2;

		init_timer2_counter();
		vd->set(&vidbm,255);
		t1 = read_timer2();
		vd->set(&vidbm,0);
		t2 = read_timer2();

		benchm.viddrv_write = (vidbm.width*vidbm.height*1193181.0)/(float)(t1-t2);
	};

	vd->blockrestore(&vidbm,bl);
	mouse_show();
}


