#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include <a/types.h>
#include "inifile.h"
#include "list.h"


char *	strcprocess(char *t,const char *src) {
	boolean quote=FALSE;
	boolean first_char=TRUE;
	boolean	exit=FALSE;
	boolean first_radix_digit=FALSE;
	char c,slash=0;
	uchar ch;

	while(!exit) {
		c=*src++;
		if (!c) {
			if (slash==2 || slash==3) *t++=ch;
			break;
		};
		switch(slash) {
		case 0:	// no slash action
			switch(c) {
			case ' ':
				if (!quote) {
					exit=TRUE;
					break;
				};
				*t++=c;
				break;
			case '"':
				if (!quote) {
					if (first_char) {
						quote=TRUE;
						break;
					};
				} else {
					quote=FALSE;
					exit=TRUE;
					break;
				};
				*t++=c;
				break;
			case '\n':
				exit=TRUE;
				break;
			case '\\':
				slash=1;
				break;
			default:
				*t++=c;
			};
			break;
		case 1: // slash
			first_radix_digit=FALSE;
			slash=0;
			switch(c) {
				case 'a': *t++='\a';	break;
				case 'b': *t++='\b';	break;
				case 'f': *t++='\f';	break;
				case 'n': *t++='\n';	break;
				case 'r': *t++='\r';	break;
				case 't': *t++='\t';	break;
				case 'v': *t++='\v';	break;
				case '\\':*t++='\\';	break;
				case '?': *t++='?';	break;
				case ''': *t++=''';	break;
				case '"': *t++='"';	break;
				case '0': case '1': case '2': case '3':
				case '4': case '5': case '6': case '7':
					  --src;
					  slash=2;	break;
				case 'x': slash=3;	break;
			};
			ch=0;
			break;
		case 2: // octal character
			if ((c&0xf8)==0x30) {
				ch=(ch<<3)+(c-'0');
				first_radix_digit=!first_radix_digit;
			} else {
				*t++=ch;
				--src;
				slash=0;
			};
			if (!first_radix_digit) { *t++=ch; slash=0; }
			break;
		case 3: // hexadecimal character
			if (isxdigit(c)) {
				ch<<=4;
				if (c<('9'+1))	ch+=(c-'0');
				else		ch+=((c&0xdf)+10-'A');
				first_radix_digit=!first_radix_digit;
			} else {
				*t++=ch;
				--src;
				slash=0;
			};
			if (!first_radix_digit) { *t++=ch; slash=0; }
			break;
		};
		first_char=FALSE;
	}

	*t = 0;

	return t;
}


boolean IniProcessValue(char *iniline,struct IniItem *ii) {
	long x;
	char *s,*t;
	int c;

	s=(char *)ii->ivalue;

	switch(ii->type & INI_TYPEMASK) {
	case INI_ULONG:
//		sscanf(iniline," %lu \n",&ii->ivalue);

		while(*iniline == ' ') ++iniline;

		if (*iniline == '$')			{ iniline+=1; ii->ivalue = strtol(iniline,&iniline,16); }
		else if (*(ushort *)iniline == 'x0')	{ iniline+=2; ii->ivalue = strtol(iniline,&iniline,16); }
//		else sscanf(iniline,"%lu",&ii->ivalue);
		else ii->ivalue = strtoul(iniline,&iniline,10);

		while(*iniline == ' ') ++iniline;
		if (*iniline++) return FALSE;		// non-terminated

		if ((ulong)(ii->ivalue)<(ulong)(ii->low))
			ii->ivalue=ii->low;
		else if ((ulong)(ii->ivalue)>(ulong)(ii->high))
			ii->ivalue=ii->high;

		ii->type |= INI_FOUND;
		break;
	case INI_SLONG:
//		sscanf(iniline," %ld \n",&ii->ivalue);

		while(*iniline == ' ') ++iniline;

		if (*iniline == '$')			{ iniline+=1; ii->ivalue = strtol(iniline,&iniline,16); }
		else if (*(ushort *)iniline == 'x0')	{ iniline+=2; ii->ivalue = strtol(iniline,&iniline,16); }
//		else sscanf(iniline,"%ld",&ii->ivalue);
		else ii->ivalue = strtol(iniline,&iniline,10);

		while(*iniline == ' ') ++iniline;
		if (*iniline++) return FALSE;		// non-terminated

		if (ii->ivalue<ii->low) ii->ivalue=ii->low;
		else if (ii->ivalue>ii->high) ii->ivalue=ii->high;
		ii->type |= INI_FOUND;
		break;
	case INI_FLOAT:
		if (sscanf(iniline," %f \n",&ii->fvalue))
			ii->type |= INI_FOUND;
		break;
	case INI_BOOL:
		s = iniline;
		while(isspace(*s)) ++s;

		c=toupper(*s);

		if (c!='Y' && c!='T' && c!='1' && !(c=='O' && toupper(s[1])=='N')) ii->ivalue = FALSE;
		else	ii->ivalue = TRUE;

		ii->type |= INI_FOUND;

		break;
	case INI_DSTRLST:
	case INI_DSTRING:
	case INI_DSTRLIT:
		if (!(s=malloc(INI_MAX_STRING))) break;
		if ((ii->type&INI_TYPEMASK) == INI_DSTRLIT) ++s;
		if ((ii->type&INI_TYPEMASK) == INI_DSTRLST) s+=sizeof(struct Node);
	case INI_STRING:
		t = strcprocess(s,iniline) + 1;

		x=(t-s)-1;
		if ((ii->type&INI_TYPEMASK) != INI_DSTRLIT)
			if (x<ii->low || x>ii->high)
				strcpy(s,(char *)(ii->deflt));
		if ((ii->type&INI_TYPEMASK) == INI_DSTRING) {
			if (ii->ivalue && ii->ivalue!=ii->deflt)
				free((void *)ii->ivalue);
			realloc(s,x+1);
			ii->ivalue=(long)s;
		} else if ((ii->type&INI_TYPEMASK) == INI_DSTRLST) {
			realloc(s-sizeof(struct Node),x+1+sizeof(struct Node));
			AddTail((struct List *)ii->ivalue,(struct Node *)(s-sizeof(struct Node)));
		} else if ((ii->type&INI_TYPEMASK) == INI_DSTRLIT) {
			realloc(s-1,x+2);
			ii->ivalue=(long)(s-1);
			s[-1]=x;
		};

		ii->type |= INI_FOUND;
		break;
	case INI_FUNCTION:
		return ((boolean (*)(char *))ii->ivalue)(iniline);
	}

	return TRUE;
}

boolean	IniProcessLine(char *line,struct IniGroup *iglist) {
	char item_name[INI_ITEM_NAME_MAX];
	struct IniItem *iit;
	unsigned int linelen;

	sscanf(line,"%s =%n",item_name,&linelen);

	iit=iglist->items;
	while(iit->name) {
		if (!stricmp(iit->name,item_name)) break;
		++iit;
	}
	if (!iit->name) return TRUE;

	return !IniProcessValue(line+linelen,iit);
}

ulong	IniProcessGroup(FILE *f,char *linebuf,struct IniGroup *iglist) {
	char group_name[INI_GROUP_NAME_MAX];
	struct IniGroup *igl;
	short c;
	long lcnt=0;
	boolean ignore_group = FALSE;

	fscanf(f,"%s \n",group_name);

	if (group_name[strlen(group_name)-1]!=']') return -1;
	else group_name[strlen(group_name)-1]='\0';

	igl=iglist;
	while(igl->name) {
		if (!stricmp(igl->name,group_name)) break;
		++igl;
	};
	if (!igl->name) ignore_group = TRUE;

	for(;;) {
		if (EOF==(c=getc(f))) break;

		if (c=='[') {
			ungetc(c,f);
			break;
		} else if (isalnum(c)) {
			ungetc(c,f);

			// read entire line into linebuf

			fgets(linebuf,512,f);

			if (!ignore_group) {
				if (*linebuf) {
					char *t = linebuf+strlen(linebuf);
					if (t[-1]=='\n') t[-1]=0;
				}

				if (IniProcessLine(linebuf,igl)) return -1;
			}
		} else if (c==';') do { c=getc(f); } while(c!=EOF && c!='\n');
		else if (c!='\n') return -1;	/* error! */

		++lcnt;
	}

	return lcnt;
}

ulong	IniProcessGroupMem(struct IniFileLine **lineptrptr,struct IniGroup *iglist) {
	char group_name[INI_GROUP_NAME_MAX];
	struct IniGroup *igl;
	int c;
	long lcnt=0;
	boolean ignore_group = FALSE;
	struct IniFileLine	*line = *lineptrptr;
	char	*t = group_name;
	char	*s;

	// find the start of a group

	while(line) {
		if (line->line[0] == '[') break;
		line = line->next;
	}

	if (!line) return 0;	// end of file

	s = line->line+1;
	while(c=*s++) if (c==']') break; else *t++ = c;
	*t=0;

	igl=iglist;
	while(igl->name) {
		if (!stricmp(igl->name,group_name)) break;
		++igl;
	};
	if (!igl->name) ignore_group = TRUE;

	while(line=line->next) {
		c = line->line[0];
		if (c=='[') {
			break;
		} else if (isalnum(c)) {
			if (!ignore_group)
				if (IniProcessLine(line->line,igl)) return -1;
		} else if (c!=';' && !isspace(c) && c) return -1;	/* error! */

		++lcnt;
	}

	*lineptrptr = line;

	return lcnt;
}

ulong	IniProcessFile(FILE *f,struct IniGroup *iglist) {
	short c;
	long lcnt=1;
	long lc;

	char *linebuf;

	if (!(linebuf = malloc(512))) return -1;

	for(;;) {
		if (EOF==(c=getc(f))) break;

		if (c=='[') {
			if (-1==(lc=IniProcessGroup(f,linebuf,iglist))) return lcnt;
			lcnt+=lc;
		} else if (c==';') do { c=getc(f); } while(c!=EOF && c!='\n');
		else if (c!='\n') return lcnt;	/* error! */

		++lcnt;
	}

	return 0;	/* successful process! */
}

ulong	IniProcessFileMem(struct IniFile *inifile,struct IniGroup *iglist) {
	uchar c;
	long lcnt=1;
	long lc;
	struct IniFileLine *ifl = inifile->first;

	while(ifl) {
		c = ifl->line[0];

		if (c=='[') {
			if (-1==(lc=IniProcessGroupMem(&ifl,iglist))) return lcnt;
			lcnt+=lc;
			continue;
		} else if (c!=';' && !isspace(c) && c) return lcnt;	/* error! */

		++lcnt;
		ifl=ifl->next;
	}

	return 0;	/* successful process! */
}

struct IniFile *IniCreateFile(void) {
	struct IniFile *inifile;

	if (!(inifile = malloc(sizeof(struct IniFile)))) return NULL;

	inifile->first = NULL;
	return inifile;
}

struct IniFile *IniReadFile(FILE *f) {
	struct IniFile *inifile;
	struct IniFileLine *last_line = NULL;
	struct IniFileLine *new_line;
	char *linebuf;
	int len;

	if (!(inifile = malloc(sizeof(struct IniFile)))) return NULL;

	inifile->first = NULL;

	if (!(linebuf = malloc(512))) goto abort1;

	// read the file in a line at a time

	while(fgets(linebuf,512,f)) {

		// seek and destroy newline

		len=strlen(linebuf);
		if (len) if (linebuf[len-1]=='\n') { linebuf[len-1]=0; --len; }

		// allocate space and link in the line

		if (!(new_line = malloc(sizeof(struct IniFileLine)+len)))
			goto abort2;

		strcpy(new_line->line,linebuf);
		new_line->prev = last_line;
		new_line->next = NULL;
		if (last_line)	last_line->next = new_line;
		else		inifile->first	= new_line;

		last_line = new_line;
	}

	free(linebuf);
	return inifile;

abort2:	free(linebuf);
abort1:	IniDestroyFile(inifile);
	return NULL;
}

boolean	IniWriteFile(FILE *f,struct IniFile *inif) {
	struct IniFileLine *ifl;
	boolean	last_was_comment = FALSE;
	boolean	last_was_blank = TRUE;
	boolean	is_blank;

	ifl = inif->first;
	while(ifl) {
		is_blank = isspace(ifl->line[0]) || !ifl->line[0];

		switch(ifl->line[0]) {
		case ';':
			if (!last_was_comment && !last_was_blank) if (fputc('\n',f)<0) return FALSE;
			if (fputs(ifl->line,f)<0) return FALSE;
			if (fputc('\n',f)<0) return FALSE;
			last_was_comment = TRUE;
			break;
		case '[':
			if (!last_was_blank) if (fputc('\n',f)<0) return FALSE;
		default:
			if (!is_blank || !last_was_blank) {
				if (fputs(ifl->line,f)<0) return FALSE;
				if (fputc('\n',f)<0) return FALSE;
				last_was_comment = FALSE;
			}
			break;
		}
		last_was_blank = is_blank;

		ifl=ifl->next;
	}

	if (fflush(f)) return FALSE;

	return TRUE;
}

void	IniDestroyFile(struct IniFile *inif) {
	struct IniFileLine *ifl,*next_ifl;

	next_ifl = inif->first;
	while(ifl=next_ifl) {
		next_ifl = ifl->next;
		free(ifl);
	}

	free(inif);
}

void	IniClearItemFoundFlags(struct IniItem *ii) {
	while(ii->name) {
		ii->type &= ~INI_FOUND;
		++ii;
	}
}

//	boolean IniAddItem(struct IniFile *inif,char *groupname,char *itemname,char *format, ...);
//
//	Adds an item to an INI file.
//
//	Returns TRUE on success, FALSE on failure.

boolean	IniAddItem(struct IniFile *inif,char *groupname,char *itemname,char *format, ...) {
	struct	IniFileLine *ifl,*last_ifl=NULL,*new_ifl;

	char	*s;
	char	new_value[256];
	ulong	new_value_len;
	va_list	val;

	va_start(val,format);
	vsprintf(new_value+64,format,val);
	va_end(val);
	new_value_len = sprintf(new_value,"%s =%s",itemname,new_value+64);

	// search for the group
	// if we don't find the group, just add the name at the end

	ifl = inif->first;
	while(ifl) {
		if (ifl->line[0]=='[') {
			s=ifl->line+1;
			while(*s!=']' && *s) ++s;
			if (!strnicmp(groupname,ifl->line+1,s-ifl->line-1)) {
				// group found... scan for item

				ifl = (last_ifl=ifl)->next;

				while(ifl) {
					if (ifl->line[0]=='[') break;

					// look for =

					s = ifl->line;
					while(*s!='=' && *s && !isspace(*s)) ++s;
					if ((*s=='=' || isspace(*s)) && s>ifl->line) {

						if (!strnicmp(ifl->line,itemname,s-ifl->line)) {

							// item found, delete the line and fall through

							if (ifl->prev) ifl->prev->next=ifl->next; else inif->first = ifl->next;
							if (ifl->next) ifl->next->prev=ifl->prev;

							break;
						}
					}

					ifl = (last_ifl = ifl)->next;

				}

				// add item line

				if (!(new_ifl = malloc(sizeof(struct IniFileLine)+new_value_len)))
					return FALSE;

				// last_ifl will always exist

				if (last_ifl->next) last_ifl->next->prev = new_ifl;
				new_ifl->next	= last_ifl->next;
				last_ifl->next	= new_ifl;
				new_ifl->prev	= last_ifl;

				strcpy(new_ifl->line,new_value);

				return TRUE;
			}
		}
		ifl = (last_ifl = ifl)->next;
	}

	// group not found, add group and item at end

	if (!(new_ifl = malloc(sizeof(struct IniFileLine)+strlen(groupname)+2)))
		return FALSE;

	if (last_ifl) {
		if (last_ifl->next) last_ifl->next->prev = new_ifl;
		new_ifl->next	= last_ifl->next;
		last_ifl->next	= new_ifl;
		new_ifl->prev	= last_ifl;
	} else {
		// hmm, empty file

		new_ifl->next = new_ifl->prev = NULL;
		inif->first = new_ifl;
	}
	sprintf(new_ifl->line,"[%s]",groupname);

	last_ifl = new_ifl;

	if (!(new_ifl = malloc(sizeof(struct IniFileLine)+new_value_len)))
		return FALSE;

	if (last_ifl->next) last_ifl->next->prev = new_ifl;
	new_ifl->next	= last_ifl->next;
	last_ifl->next	= new_ifl;
	new_ifl->prev	= last_ifl;

	strcpy(new_ifl->line,new_value);

	return TRUE;
}

