#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <dos.h>
#include <malloc.h>

#include "loadlib.h"
#include "io.h"
#include "request.h"

extern ushort _Envseg;
extern void *_Envptr;
extern void *_STACKLOW;

struct PharLapHeader {
	char	magic_m;	// should be 'M'
	char	magic_q;	// should be 'Q' for PharLap 386 Relocatable
	ushort	length_left;	// length in bytes at last sector
	ushort	sectors;	// length in sectors
	ushort	relo_items;	// relocation items
	ushort	prog_pos;	// start of program position
	ushort	memory_pages;	// # pages needed for BSS+STACK
	ushort	maxalloc;	// maximum paragraphs
	ulong	stack_ptr;	// stack ptr (?)
	ushort	checksum;	// checksum (not used)
	ulong	inipc;		// initial PC
	ushort	first_reloc;	// offset in file of first reloc
	ushort	ovl_num;	// overlay #
};

boolean load_pharlap(char *n,struct ExeInfo *ei) {
	FILE *exe;
	void *progmem;
	void *relocmem;
	ulong *relocs;
	struct PharLapHeader plh;
	ulong filesize,imagesize,imagestrt,progsize;

	if (!(exe=VPfopen(n,"rb"))) return FALSE;

	// read PharLap header into memory

	if (1!=fread(&plh,sizeof plh,1,exe)) goto abort1;

	// check to make sure it is a PharLap executable

	if (plh.magic_m!='M' || plh.magic_q!='Q') goto abort1;

	// allocate memory for executable

	filesize=(plh.sectors<<9)+plh.length_left-512;

	imagestrt=plh.prog_pos<<4;
	imagesize=filesize-imagestrt;
	progsize=imagesize+(((long)plh.memory_pages)<<12);

	if (!(progmem =malloc(progsize		))) goto abort1;
	if (!(relocmem=malloc(plh.relo_items<<2	))) goto abort2;
	relocs=relocmem;

	memset(progmem,0,progsize);

	// read relocs & program

	fseek(exe,plh.first_reloc,SEEK_SET);
	if (1!=fread(relocmem,plh.relo_items<<2	,1,exe)) goto abort3;

	fseek(exe,imagestrt,SEEK_SET);
	if (1!=fread(progmem ,imagesize		,1,exe)) goto abort3;

	// apply relocs to program
	//
	// pharlap relocatable executables load at 0x80000000

	{
		ulong relocnt=plh.relo_items;

		do {
			*(ulong *)((char *)progmem + *relocs++ - 0x80000000) += (ulong)progmem;
		} while(--relocnt);
	}

	free(relocmem);	// unload relocs from memory

	// return values

	ei->start  = (void(*)(void)) ((char *)progmem+plh.inipc);
	ei->memory = progmem;

	fclose(exe);

	return TRUE;
abort3:
	free(relocmem);
abort2:
	free(progmem);
abort1:
	fclose(exe);
	return FALSE;
};

void	unload_pharlap(struct ExeInfo *ei) {
	free(ei->memory);
	ei->memory=NULL;
}

long run_pharlap_asm(void(*)(void),ushort,ushort,void *,void *,ulong);
#pragma aux run_pharlap_asm="push ebp","call esi","pop ebp" modify exact [eax ebx ecx edx esi edi] \
	parm [esi] [bx] [cx] [edx] [eax] [edi] value [eax];

long	run_pharlap(struct ExeInfo *ei) {
	return run_pharlap_asm(ei->start,_psp,_Envseg,_Envptr,_STACKLOW,0);
}

void	revector_pharlap(struct ExeInfo *ei,ulong *reloclst) {
	ulong reloc_vectors;
	ulong reloc_data;
	ulong *extrlst=((ulong **)ei->start)[-1];

	reloc_vectors	=min(reloclst[0],extrlst[0]);
	reloc_data	=min(reloclst[1],extrlst[1]);

	memcpy(((char *)(extrlst+2))+5*extrlst[0],((char *)(reloclst+2))+4*reloclst[0],reloc_data<<2);

	reloclst+=2;
	extrlst +=2;

	do {
		extrlst=(ulong *)((char *)extrlst + 1);
		*extrlst=*reloclst++-(((ulong)(char *)extrlst)+4);
		++extrlst;
	} while(--reloc_vectors);
}


////////////////////////////// EXTENSION ROUTINES ////////////////////////////////


extern ulong VPpatchentry[];


boolean	run_extension(char *extname) {
	struct ExeInfo ei;

	if (load_pharlap(extname,&ei)) {
		revector_pharlap(&ei,VPpatchentry);
		run_pharlap(&ei);
		unload_pharlap(&ei);
		return TRUE;
	} else	return FALSE;
}


static struct List ExtList={NULL};
static ulong ExtCount = 0;


//	ext = load_extension(extname)
//
//	Load, initialize, and link in an extension

struct Extension *load_extension(char *extname) {
	struct Extension *ext;

	if (!(ext=(struct Extension *)malloc(sizeof(struct Extension)))) return NULL;
	memset(ext,0,sizeof(struct Extension));

	if (!ExtList.next) NewList(&ExtList);

	if (load_pharlap(extname,&ext->ei)) {
		revector_pharlap(&ext->ei,VPpatchentry);

		// call the STARTUP point... if it fails, return NULL

		ext->entry=(ExtEntry)run_pharlap_asm(ext->ei.start,_psp,_Envseg,_Envptr,_STACKLOW,EC_STARTUP);
		switch((ulong)ext->entry) {
		case 0:	// load aborted by extension
			free(ext);
			return NULL;
		case 1:	// extension loaded successfully but self-unloaded because no reason to load (say, no hware)
			free(ext);
			return (void *)-1L;
		default:
			// call the STARTUP1 point

			if (ext->entry(ext,EC_STARTUP1,0,0)==TRUE) {
				free(ext);
				return 0;
			}

			// extension entry point has been returned - link into list

			AddTail(&ExtList,&ext->node);
			++ExtCount;

			return ext;
		};
	};
	return NULL;
}


//	failure = unload_extension(extension)
//
//	Deinitialize, unlink, and remove an extension from memory


boolean	unload_extension(struct Extension *ext) {

	// check the usecount - if >0, can't unload

	if (ext->usecount) return TRUE;

	// call the extension with the command EC_CLEANUP, and see if it concurs

	if (ext->entry(ext,EC_CLEANUP,0,0)) return TRUE;	// extension doesn't want to go

	// extension has now completely shut down... free and unlink

	Remove(&ext->node);
	unload_pharlap(&ext->ei);

	--ExtCount;

	return FALSE;
}


//	failure = announce_to_extensions(extension_list,command,xval,yval)
//
//	Call all extensions on a provided list and return TRUE if any of them return non-zero
//
//	This function will call ALL the extensions on the list, regardless of whether any of them
//	fail.
//
//	This function can only be used with the <user> list in the Extension structure, since the
//	global list is the _second_ entry,

boolean	announce_to_extensions(struct List *extension_list,enum ExtCommand command,long xval,long yval) {
	struct Node *np,*next;
	boolean	failed;

	np=extension_list->prev;

	while(next=np->prev) {
		if ( ((struct Extension *)np)->entry((struct Extension *)np,command,xval,yval) )
			failed=TRUE;
		np=next;
	};

	return failed;
}

//	unload_extensions(extension_list)
//
//	Unload all extensions on a provided list.
//
//	This function can only be used with the <user> list in the Extension structure, since the
//	global list is the _second_ entry,

void	unload_extensions(struct List *extension_list) {
	struct Node *np,*next;

	np=extension_list->prev;

	while(next=np->prev) {
		unload_extension((struct Extension *)np);
		np=next;
	};
}



//////////////////// EXTENSIONS DIALOGUE BOX /////////////////////

static void ExtensionListDisplayIndex(struct Listbox *lb,struct Extension *entry,int dindex,int highlight) {
	color	backg;
	coord	y;

	y=lb->y+lb->textheight*dindex;
	backg=highlight ? ST1 : ST2;

	vd->bar(lb->destbm,lb->x,y,lb->width,lb->textheight,backg);
	if (entry)
		SysTextLim(lb->destbm,lb->x,y,entry->name ? entry->name : "< no name >",highlight ? ST3 : ST0,backg,lb->cwidth);
//		SysTextPrintf(lb->destbm,lb->x,y,highlight ? ST3 : ST0,backg,"%08lx",entry);
}

//#define	offsetof(type,member) (&((type *)0)->member)

void	ExtensionDisplayInfoList(void) {
	struct ListRequester *lr;

	if (!(lr = MakeListRequester(40,10,"Extensions"))) return;

	lr->zones[1].text = "Setup";

	if (lr->tbox->table = malloc(sizeof(char *)*(ExtCount+1))) {
		struct Node *n = ExtList.prev,*n2;
		char **tp = lr->tbox->table+1;
		
		while(n2 = n->prev) {
			*tp++ = (char *)n - offsetof(struct Extension,node);
			n = n2;
		}
		lr->tbox->table[0] = (char *)ExtCount;

		TextboxCreate(lr->tbox,TRUE);
		lr->lbox->displayIndex = &ExtensionListDisplayIndex;
		if (OpenListRequesterAutosize(lr)) {
			mouse_coord	mx,my,mb;
			ulong		msg;
			boolean		cont=TRUE;
			boolean		okay = FALSE;

			do {
				mouse_read2(&mx,&my,&mb);
				if (mb & 1) {
					msg=CheckRequester(lr->req,mx,my,mb);
					mouse_wait_unkey();
					switch(msg) {
					case RZoneMsg(0,0):	/* Ok */
						okay = TRUE;
					case RZoneMsg(0,1):	/* Cancel */
						cont=FALSE;
						break;
//					case REQM_LISTBOX+0:
//						break;
					};
				};
			} while(cont);

			CloseRequester(lr->req);
		}
		free(lr->tbox->table);
	}

	DisposeListRequester(lr);
}

