////////////////////////////////////////////////////
//
// VGAPaint 386 - polygon.c
//
// Convex polygon routine.
//


#include "video.h"
#include "vgapaint.h"
#include "polygon.h"
#include <math.h>

#if 0
struct PolyConvexEdge {
	int	x;		// x-coordinate
	int	idx;		// integer displacement per y change
	int	disc;		// discriminant for fractional changes
	int	dlimit;		// discriminant limit
	int	yinc;		// discriminant change per y change
	char	inc;		// increment
};
#endif

struct PolyConvexEdge {
	double x,dx;
	long i;
};

static int delete(struct PolyConvexEdge *edge,long i,int nact);
static int insert(struct PolyVertex *point,struct PolyConvexEdge *edge,long i,long y,int nact,int n);
static void PolySort(struct PolyVertex *point,int *ind,long left,long right);
static void PolySort2(struct PolyConvexEdge *edge,long left,long right);


#if 0
		if (!pointListInit()) {
			if (fillmode == FILL_STRETCH) {
				if (!vsd->init(winbm,polybx1,polyby1,&brush_sprite.imgbm,polybx2+1-polybx1,polyby2+1-polyby1))
					goto failed;
				polystate=TRUE;
				PolygonEngine(colr);
				vsd->map_init();
			};
			polystate=FALSE;
			PolygonEngine(colr);
			if (fillmode == FILL_STRETCH)
				vsd->dispose();
		};
#endif

static void VPpolygon_convex_rec_hline(BMP dst,coord x1,coord x2,coord y,color c) {
	vsd->rec_hline(x1,x2,y);
}

static void VPpolygon_convex_map_hline(BMP dst,coord x1,coord x2,coord y,color c) {
	vsd->map_hline(dst,&brush_sprite.imgbm,x1,x2,y);
}

static void VPpolygon_convex_tile_hline(BMP dst,coord x1,coord x2,coord y,color c) {
	vd->tile(&brush_sprite.imgbm,dst,x1,y,x2+1-x1,1,vposx+brush_ox+x1,vposy+brush_oy+y);
}

boolean	VPpolygon_convex(BMP dst,struct PolyVertex *point,int n,color colr) {
	boolean ret;
	coord x1,y1,x2,y2;
	int i;

	// scan the point list for bounds

	x1 = y1 = point[0].x;
	x2 = y2 = point[0].y;

	for(i=1;i<n;i++) {
		if (point[i].x < x1) x1 = point[i].x;
		if (point[i].x > x2) x2 = point[i].x;
		if (point[i].y < y1) y1 = point[i].y;
		if (point[i].y > y2) y2 = point[i].y;
	}

	TAP_RECT(x1,y1);
	TAP_RECT(x2,y2);

	switch(fillmode) {
	case FILL_STRETCH:
		if (!vsd->init(winbm,x1,y1,&brush_sprite.imgbm,x2+1-x1,y2+1-y1)) return FALSE;
		if (!vid_polygon_convex(dst,point,n,colr,VPpolygon_convex_rec_hline)) {
			vsd->dispose();
			return FALSE;
		}
		vsd->map_init();
		ret = vid_polygon_convex(dst,point,n,colr,VPpolygon_convex_map_hline);
		vsd->dispose();
		return ret;
	case FILL_TILE:		return vid_polygon_convex(dst,point,n,colr,VPpolygon_convex_tile_hline);
	default:		return vid_polygon_convex(dst,point,n,colr,vd->hline);
	}
}

boolean vid_polygon_convex(BMP dst,struct PolyVertex *point,int n,color colr,void (*hline)(BMP dst,coord x1,coord x2,coord y,color c)) {
	long xoff,yoff;
	long k,y0,y1,i,j;
	int	nact=0;
	coord xl,xr;
	coord y;
	struct PolyConvexEdge	*edge;
	int			*ind;

	/////////////////////

	// allocate memory for the edges and index table

	if (!(edge = malloc(sizeof(struct PolyConvexEdge)*n))) return FALSE;
	if (!(ind = malloc(sizeof(int)*n))) {
		free(edge);
		return FALSE;
	}

	xoff=vposx+brush_ox;
	yoff=vposy+brush_oy;

	for(k=0;k<n;k++) ind[k]=k;

	PolySort(point,ind,0,n-1);

	k=0;
	y0=max(dst->y1,point[ind[0]].y);
	y1=min(dst->y2-1,point[ind[n-1]].y-1);
	for(y=y0;y<=y1;y++) {
		while (k<n && point[i=ind[k]].y <= y) {
			j=(i>0)?(i-1):(n-1);
			if (point[j].y<=y) nact = delete(edge,j,nact);
			else if (point[j].y>=y) nact = insert(point,edge,j,y,nact,n);

			j=(i<n-1)?(i+1):0;

			if (point[j].y<=y) nact = delete(edge,i,nact);
			else if (point[j].y>=y) nact = insert(point,edge,i,y,nact,n);
			k++;
		};
		PolySort2(edge,0,nact-1);

		for (j=0;j<nact;j+=2) {
			xl=ceil(edge[j].x-.5);
			if (xl<dst->x1) xl=dst->x1;
			xr=floor(edge[j+1].x-.5);
			if (xr>=dst->x2) xr=dst->x2-1;
			if (xl<=xr)
#if 0
				switch(fillmode) {
				case FILL_TILE:
					vd->tile(&brush_sprite.imgbm,dst,xl,y,xr+1-xl,1,xoff+xl,yoff+y);
					break;
				case FILL_STRETCH:
					if (polystate)
						vsd->rec_hline(xl,xr,y);
					else
						vsd->map_hline(dst,&brush_sprite.imgbm,xl,xr,y);
					break;
				default:
					vd->hline(dst,xl,xr,y,colr);
				};
#endif
				hline(dst,xl,xr,y,colr);
		};
		for (j=0;j<nact;j++)
			edge[j].x+=edge[j].dx;
	}

	free(ind);
	free(edge);

	return TRUE;
}

static int delete(struct PolyConvexEdge *edge,long i,int nact) {
	long j,k;

	for (j=0;j<nact;j++)
		if (edge[j].i==i) {
			nact--;
			for (k=j;k<nact;k++)
				edge[k]=edge[k+1];
			return nact;
		}

	return nact;
}

static int insert(struct PolyVertex *point,struct PolyConvexEdge *edge,long i,long y,int nact,int n) {
	long j,p,q;
	double dxt;

	j=(i<n-1)?(i+1):0;
	if (point[i].y < point[j].y) { p=i; q=j; }
	else { p=j; q=i; };

	edge[nact].dx=dxt=(double)(point[q].x - point[p].x)/(double)(point[q].y - point[p].y);
	edge[nact].x=dxt * (double)((double)y+.5-(double)point[p].y)+(double)point[p].x;
	edge[nact].i=i;

	return nact+1;
}

#define Pswap(x,y)  { temp=ind[x]; ind[x]=ind[y]; ind[y]=temp; }
#define Pswap2(x,y) { temp=edge[x]; edge[x]=edge[y]; edge[y]=temp; }

/* PolySort - Sort w/quicksort */

static void PolySort(struct PolyVertex *point,int *ind,long left,long right) {
	long temp,last,i;

	if (left>=right) return;
	Pswap(left,(left+right)>>1);
	last=left;
	for (i=left+1;i<=right;i++)
		if (point[ind[i]].y<point[ind[left]].y) {
			++last; /* do not double-increment */
			Pswap(last,i);
		};
	Pswap(left,last);
	PolySort(point,ind,left,last-1);
	PolySort(point,ind,last+1,right);
}

static void PolySort2(struct PolyConvexEdge *edge,long left,long right) {
	long last,i;
	struct PolyConvexEdge temp;

	if (left>=right) return;
	Pswap2(left,(left+right)>>1);
	last=left;
	for (i=left+1;i<=right;i++)
		if (edge[i].x<edge[left].x) {
			++last; /* do not double-increment */
			Pswap2(last,i);
		};
	Pswap2(left,last);
	PolySort2(edge,left,last-1);
	PolySort2(edge,last+1,right);
}
