#ifndef	VGAPAINT_REQUEST_H
#define	VGAPAINT_REQUEST_H

#include "vgapaint.h"
#include "listbox.h"
#include "mouse.h"

/* 3d color translate */
#define	ctran(x)	(((color *)&col)[x])

struct Text {
	char *s;
	coord x,y;
	color fore,back;	/* in 3d colors */
};

#define	REQM_INTERNAL	(0x00000000)
#define	REQM_NOTHING	(REQM_INTERNAL+0)
#define	REQM_DRAGGED	(REQM_INTERNAL+1)

#define	REQM_ZONE	(0x01000000)
#define	REQM_FIELD	(0x02000000)
#define	REQM_LISTBOX	(0x03000000)
#define	REQM_SLIDER	(0x04000000)

#define	REQM_TYPE	(0xFF000000)
#define	REQM_SHIFT	(24)
#define	REQM_MESSAGE	(0x00FFFFFF)

#define	RZoneMsg(zr,z)	(((long)(zr)<<16)+(z)+REQM_ZONE)
#define	RZoneMR(msg)	ZoneMR((msg)&REQM_MESSAGE)
#define	RZoneMZ(msg)	ZoneMZ((msg)&REQM_MESSAGE)
#define	RMsgIsZone(msg)	(((msg)&REQM_TYPE)==REQM_ZONE)

// REQ_BORDER		-> 3D border the requester
// REQ_NOSAVEBACK	-> Don't try to save background
// REQ_NODRAG		-> Don't allow dragging (but still send message)
// REQ_POPCEN		-> Pop up centered on mouse pointer
// REQ_FORCECOL		-> Force colors to standard
// REQ_COORDBIAS	-> Bias requester so that titlebar is above (0,0)
// REQ_SLDREPHIT	-> Repeat-hit slider gadgets
// REQ_SYSTITLE		-> Use system font to render titles
// REQ_NOINHIBITCC	-> Don't inhibit colorcycling

// REQ_COLCHANGED	-> Internal: Colors were changed
// REQ_HIDDEN		-> Internal: Requester is hidden

#define	REQ_BORDER	(0x00000001)
#define	REQ_NOSAVEBACK	(0x00000002)
#define	REQ_NODRAG	(0x00000004)
#define	REQ_POPCEN	(0x00000008)
#define	REQ_FORCECOL	(0x00000010)
#define	REQ_COORDBIAS	(0x00000020)
#define	REQ_SLDREPHIT	(0x00000040)
#define	REQ_SYSTITLE	(0x00000080)
#define	REQ_NOINHIBITCC	(0x00000100)

#define	REQ_COLCHANGED	(0x00010000)
#define	REQ_HIDDEN	(0x00020000)

struct Requester {
	long		curx,cury;	/* current position */
	long		width,height;	/* size */
	ulong		flags;		/* flags */
	char		*title;		/* name of requester */
	color		backfill;	/* backfill color */
	struct Listbox	*lboxes;	/* Listbox array */
	struct Field	*fields;	/* Field array */
	struct ZoneRange *zones;	/* Zone range lists */
	struct Slider	*sliders;	/* Slider list */
	struct Rectangle *pblist;	/* Positive 3D box list */
	struct Rectangle *nblist;	/* Negative 3D box list */
	struct Text	*tlist;		/* Text list */

	// working registers 

	struct BitMap	reqbm;		/* all rendering takes place here */
	struct BitMap	*old_winbm;	/* winbm save (kludge) */
	uchar		*savepal;	// palette memory save
	struct BitMap	*old_zonebm;	/* zonebm save */
	BLOCK		saveback;	/* background save */
	long		savewidth,saveheight;
};

struct	StatusBar {
	struct Requester *req;
	long	maximum;
	long	atpoint;
	coord	curpixels;
};

struct	ListRequester {
	struct Requester *req;
	struct Textbox	 *tbox;
	struct Listbox	 *lbox;
	struct Zone	 *zones;
	struct ZoneRange *zr;
	struct Slider	 *sld;
	struct Rectangle *neg;
	int	width;
	int	height;
};

struct	SliderRequester {
	struct Requester req;
	struct Slider sld;
};

void	ClipRequesterPos(struct Requester *,long *,long *);
boolean	OpenRequester(struct Requester *,long,long);
void	CloseRequester(struct Requester *);
ulong	CheckRequester(struct Requester *,mouse_coord x,mouse_coord y,mouse_coord b);
void	DragRequester(struct Requester *);
void	MoveRequester(struct Requester *,long,long);
void	RenderRequester(struct Requester *);
void	HideRequester(struct Requester *req);
void	ShowRequester(struct Requester *req);

long	RequestSlider(char *,long,ulong,long);
struct SliderRequester *MakeSliderRequester(long width);
void	DisposeSliderRequester(struct SliderRequester *sldr);

short	RequestBoolean(char *,char *,char *,char *,char *, ...);

struct	StatusBar *	CreateStatusBar(long maxval,char *format, ...);
void			ModifyStatusBar(struct StatusBar *,long);
void			DestroyStatusBar(struct StatusBar *);

struct	ListRequester *	MakeListRequester(long xsiz,long ysiz,char *name);
void	ResizeListRequester(struct ListRequester *lr,coord xsiz,coord ysiz);
long	ActivateListRequesterAutosize(struct ListRequester *lr);
boolean	OpenListRequesterAutosize(struct ListRequester *lr);
long	ActivateListRequester(struct ListRequester *lr);
long	ActivateListRequester1(struct ListRequester *lr);
void	DisposeListRequester(struct ListRequester *lr);

void	Render3DList(BMP,struct Rectangle *,boolean);
void	RenderTextList(BMP,struct Text *);

#endif
