#ifndef	VGAPAINT_TOOLICON_H
#define VGAPAINT_TOOLICON_H

#include "vgapaint.h"
#include "video.h"
#include "list.h"

#ifdef	VGAPAINT_TOOLICON_C
#define	EXTERN
#define	INIT(x) =x
#else
#define	EXTERN extern
#define INIT(x)
#endif

void	clipOff(void);
void	clipOn(void);
void	clipReset(void);
void	set_tools_mode(int);
void	remake_clip_bitmaps(void);
void	SetTitle(char *f, ...);
void	display_tools(void);
void	draw_tools(void);
void	erase_tools(void);
void	restore_tools(void);
void	restore_screen(void);
void	reblit_screen(void);
void	syncbitmap(void);
void	Undo(void);
uchar	request_color(uchar *palette,uchar red,uchar grn,uchar blu,short orgcol);
//ulong	rmsdif(uchar cnum1,uchar cnum2);
ulong	luminance(uchar cnum);
void	choose_colors(uchar *palette);
void	repick_colors(void);
void	reset_colors(void);
void	restore_colors(void);

boolean	pick_point(coord *x,coord *y);

void	check_scroll(void);
void	scroll_speed_req(void);

void	scroll_viewport(long xdelta,long ydelta);
void	redraw_picture(void);
boolean	resize_picture(long xs,long ys,short action);
boolean	translate_picture_start(struct BitMap *bmtemp,long x,long y);
void	translate_picture_end(struct BitMap *bmold);

void	pic_flip_x(void);
void	pic_flip_y(void);
boolean	pic_rotate_right(void);

void	ZoomOff(void);
void	ZoomOn(coord zx,coord zy);
void	ZoomHandler(void);
void	ZoomRefreshHandler(void);

void	Overview(void);

void	dragscreen(void);

void	SetSystemFont(struct Font *);
void	SysText(BMP,coord,coord,char *,color,color);
void	SysTextPrintf(BMP bm,coord x1,coord y1,color f,color b,char *s, ...);
void	SysTextLim(BMP,coord,coord,char *,color,color,ulong lim);
void	SysTextTrans(BMP,coord,coord,char *,color);
ulong	TextLength(struct Font *,char *);

boolean	AddVideoDriver(struct VideoDriver *);
boolean	RemVideoDriver(struct VideoDriver *);
void	InitBuiltinVideoDrivers(void);
void	SelectVideoDriver(void);
int	SwitchVideoDriver(struct VideoDriver *vdvr,boolean report);
boolean	SwitchBuiltinVideoDriver(long vd_code,boolean report);

void	InstallHarderrHandler(void);
void	DeinstallHarderrHandler(void);


enum	{
	VDB_Mode13=0,
	VDB_Mode13Buffered,
	VDB_Mode13X320x400,
	VDB_Mode13X360x400,
	VDB_Mode13X320x480,
	VDB_Mode13X360x480,
	VDB_Mode13H,
	VDB_Mode640x400,
	VDB_Mode640x480,
	VDB_Mode800x600,
	VDB_Mode1024x768,
	VDB_Mode1280x1024
};

enum VideoDriverSwitchFail {
	VDSF_SWITCH_OK		= 0,	// switch was ok
	VDSF_SWITCH_FAILED	= 1,	// switch failed, using VGA driver
	VDSF_SWITCH_NOVIDEO	= 2,	// switch failed bigtime, no driver!
	VDSF_SWITCH_UNKNOWN	= 3,	// switch failed, no change
};

struct	VideoDriverInitCall {
	enum VD_InitCode	vdic;
	long			x;
	long			y;
};

struct	VideoDriver {
	struct	Node	node;		// linked list
	char	*name;			// no more than 30 chars
	struct	VDev		*vd;	// video device pointer
	struct	VideoDriverInitCall	*vdiclist;	// video device init call list
	ulong	flags;			// flags
	ushort	maxcol;			// max colors
	ushort	VDsize;			// size of VD video structure
	ushort	VDrvSize;		// size of VideoDriver structure
	uchar	x_aspect;		// 1=320, 2=640, etc. -- should be truncated down
	uchar	y_aspect;		// 1=200, 2=400, etc. -- should be truncated down
	ulong	x_ratio,y_ratio;	// default ratios
	char	*tag;			// shorthand tag
};

EXTERN struct VideoDriver *vdrv_cur;

#define	VDRV_SIZE	(sizeof(struct VideoDriver))

enum {
	VDF_NONSTD	=0x00000001,	// not a standard VGA/VESA mode
	VDF_OKAY	=0x00000002,	// usually okay (but not always) on most VGA's (Mode X, etc.)
	VDF_HACK	=0x00000004,	// real hack, may not work on some VGA's, and Windows will probably puke at it
	VDF_VESA	=0x00000008,	// this driver requires VESA support
	VDF_FONTMASK	=0x000000F0,
	VDF_FONT6x8	=0x00000000,
	VDF_FONT8x8	=0x00000010,
	VDF_FONT8x12	=0x00000020,
	VDF_FONT12x16	=0x00000030,
	VDF_VHADJUST	=0x00000100,	// enable "Screen Density" adjustment
	VDF_RATIO	=0x00000200,	// ratios included
	VDF_NOBUFFER	=0x00000400,	// this driver is NOT screen buffered
	VDF_HASTAG	=0x00000800,
};

#define	VDEV_SIZE	(sizeof(struct VDev)>>2)

enum ToolBarMode {
	TOOLMODE_RIGHT,
	TOOLMODE_TOP
};

///////////////// Tool layout /////////////////////

EXTERN struct ToolPosition {
	coord	b_width;	// titlebar width
	coord	b_height;	// titlebar height/bottom
	coord	t_x1;		// toolbar left
	coord	t_y1;		// toolbar top
	coord	t_width; 	// toolbar width
	coord	t_height;	// toolbar height
	coord	i_width;	// icon width
	coord	i_height;	// icon height
	coord	c_width;	// palette color width
	coord	c_height;	// palette color height
	coord	tools_xmult;	// x magnification for tools
	coord	tools_ymult;	// y magnification for tools
	coord	tools_xpoint;	// pixels per point (tools)
	coord	tools_ypoint;	// pixels per point (tools)
	boolean	use_hires;
} toolpos;

#undef	EXTERN
#undef	INIT

#endif
