#define	VGAPAINT_VIDVESA_C
#include "vidvesa.h"
#include "io.h"
#include "toolicon.h"

/////////////// WARNING //////////////////

// This module is compiled with -zu so that
// the stack is not the same selector as the
// data seg.  NO AUTOMATIC ARRAYS REFERENCED
// BY POINTERS.

static long	VESA_Width,VESA_Height,VESA_Bpr;
static short	VESA_Mode;
static uchar *	VESA_Window;
static ulong	VESA_Window_Size;
static ulong	VESA_Window_Mask;
static short	VESA_Window_Bits;
static ushort	VESA_Window_ID;
static short	VESA_Bank_Shift;

static uchar *	VESA_Read_Window;
static ushort	VESA_Read_Window_ID;

static void	(*VESA_Display_Win_Func)(void);

static boolean	VESA_Use_Linear = FALSE;
static ulong	VESA_Physical_Window=NULL;
static uchar	*VESA_Linear_Window=NULL;
static ulong	VESA_Linear_Window_Size=0;
static ushort	VESA_Selector=0;

extern void VESA_Display_Window_16(void);

////////////////////

static char *VESA_Mode_Fault_Messages[]={
	"Mode not supported.",
	"Mode is not a graphics mode.",
	"Mode does not have a Read Window.",
	"Mode does not have a Write Window.",
	"Mode has a very, very small window.",
	"Mode has a weird window size.",
};

void	VESA_Get_Info(ushort mode,struct VESAModeInfo *vmi) {
	ushort desc;
	void *mptr;
	static struct RealInt rregs;

	if (!(mptr=DosMemoryAllocate(17,&desc))) {
		memset(vmi,0,256);
		return;
	};

	memset(mptr,0,256);

	rregs.w.ax=0x4f01;
	rregs.w.cx=mode;
	rregs.es=(ulong)mptr>>4;
	rregs.w.di=(ulong)mptr & 15;
	DPMIRealInt(0x10,&rregs);

	memcpy(vmi,mptr,256);
	DosMemoryFree(desc);

}

///////////////////////

struct VESA_Scanned_Mode_Info {
	struct VideoDriver		vd_struct;
	struct VideoDriverInitCall	initcall[2];
	char				modename[40];
};

boolean	VESA_Scan_Modes(void) {
	ushort desc;
	void *mptr;
	struct RealInt r;
	ushort *modes_list;
	struct VESAModeInfo *vmi;
	struct VESA_Scanned_Mode_Info *vsmi;
	ulong	flags;

	if (!VESA_Selector)
		if (!(VESA_Selector = DPMIAllocateSelector())) return FALSE;

	if (!(mptr=DosMemoryAllocate(32+17,&desc)))
		return FALSE;

	memset(mptr,0,512);
	*(ulong *)mptr = '2EBV';

	// query VESA for the mode list

	r.x.eax = 0x4f00;
	r.es = ((ulong)mptr) >> 4;
	r.w.di = ((ulong)mptr) & 15;
	DPMIRealInt(0x10,&r);

	if (r.w.ax != 0x004f) goto failure;

	modes_list = (ushort *)((ulong)(((ushort *)mptr)[7]) + ((ulong)(((ushort *)mptr)[8])<<4));

	while(*modes_list != 0xffff) {
		// get VESAModeInfo for this mode

		memset(&r,0,sizeof r);
		r.x.eax	= 0x4f01;
		r.es	= (((ulong)mptr) >> 4)+32;
		r.w.di	= ((ulong)mptr) & 15;
		r.w.cx	= *modes_list++;
		memset((uchar *)mptr+512,0,17*16);

		DPMIRealInt(0x10,&r);
		if (r.w.ax != 0x004f) continue;

		// feed it to check function

		vmi = (struct VESAModeInfo *)((uchar *)mptr + 512);

//printf("VESA_Check_Mode_Acceptable(mode %03x) [1]\n",modes_list[-1]); getch();

		if (!VESA_Check_Mode_Acceptable(vmi,FALSE)) {

//printf("VESA_Check_Mode_Acceptable(mode %03x) [2]\n",modes_list[-1]); getch();

			if (vsmi = malloc(sizeof(struct VESA_Scanned_Mode_Info))) {

				flags = VDF_VESA + VDF_RATIO;

				if ((vmi->mode_attr & VMI_MODE_LINEAR_ABLE) && !opt.noVESA_LFB)
					flags |= VDF_NOBUFFER;

				if	(vmi->hres>=1024 && vmi->vres>=768)	flags += VDF_FONT12x16;
				else if (vmi->hres>= 800 && vmi->vres>=600)	flags += VDF_FONT8x12;
				else if (vmi->hres>= 640 && vmi->vres>=400)	flags += VDF_FONT8x8;
				else						flags += VDF_FONT6x8;

				sprintf(vsmi->modename,"VESA %4dx%4dx256 (%03x%s)",vmi->hres,vmi->vres,modes_list[-1],
					flags&VDF_NOBUFFER ? "-LFB" : "");
				vsmi->initcall[0].vdic	= VD_SETSIZE;
				vsmi->initcall[0].x	= modes_list[-1];
				vsmi->initcall[0].y	= 0;
				vsmi->initcall[1].vdic	= NULL;
				vsmi->initcall[1].x	= 0;
				vsmi->initcall[1].y	= 0;
				vsmi->vd_struct.name	= vsmi->modename;
				vsmi->vd_struct.vd	= &vdvesa;
				vsmi->vd_struct.vdiclist= vsmi->initcall;
				vsmi->vd_struct.flags	= flags;
				vsmi->vd_struct.maxcol	= 256;
				vsmi->vd_struct.VDsize	= VDEV_SIZE;
				vsmi->vd_struct.VDrvSize= VDRV_SIZE;
				vsmi->vd_struct.x_aspect= vmi->hres/320;
				vsmi->vd_struct.y_aspect= vmi->vres/200;
				vsmi->vd_struct.x_ratio	= 0x10000;
				vsmi->vd_struct.y_ratio	= 0x10000;
				AddVideoDriver(&vsmi->vd_struct);
			}
		}

//	printf("VESA_Check_Mode_Acceptable(mode %03x) [3]\n",modes_list[-1]); getch();

	}
	DosMemoryFree(desc);
	return TRUE;

failure:
	DosMemoryFree(desc);
	return FALSE;
}

//	enum VESAModeCheck (int) fault = VESA_Check_Mode_Acceptable(struct VESAModeInfo *vmi,
//		boolean set_mode);
//
//	Check the VESA mode information returned from the video BIOS for acceptability.
//
//	Returns a non-zero result if the mode is screwy for some reason.
//
//	If <set_mode> is TRUE, global variables will be initialized to set up for that
//	particular VESA mode.

int	VESA_Check_Mode_Acceptable(struct VESAModeInfo *vmi,boolean set_mode) {
	short window=0,rwindow=0;
	ushort vms;
	struct RealInt r;

	if (!(vmi->mode_attr & VMI_MODE_SUPPORTED)) return VMCHK_NOT_SUPPORTED;
	if (!(vmi->mode_attr & VMI_MODE_GRAPHICS)) return VMCHK_NOT_GRAPHICS;

	// width,height cannot be below (320,200)

	if (vmi->hres<320 || vmi->vres<200) return VMCHK_TOO_SMALL;

	if (set_mode) {
		VESA_Width= vdvesa_specs.minx=vdvesa_specs.maxx=vdvesa_specs.o_vidx=vdvesa_specs.o_maxx = vmi->hres;
		VESA_Height=vdvesa_specs.miny=vdvesa_specs.maxy=vdvesa_specs.o_vidy=vdvesa_specs.o_maxy = vmi->vres;
	}

	// must be 8 bitsperpixel

	if (vmi->bpp != 8) return VMCHK_WRONG_DEPTH;

	// check to see if VESA VBE 2.0 32-bit protected mode functions are supported

	if (set_mode) {

		if (opt.noVESA32bit) {
			r.w.ax = 0;
		} else {
			r.x.eax = 0x4f0a;
			r.b.bl = 0;
			DPMIRealInt(0x10,&r);
		}

		if (r.w.ax == 0x004f) {
			ushort *tp;
			ushort *pp;

			tp = (ushort *)(void *)(((ulong)r.es<<4)+(ulong)r.w.di);

			VESA_Display_Win_Func = (void *)((char *)tp + tp[0]);

			if (tp[6]) {
				pp = (void *)((char *)tp + tp[6]);
				while(*pp++ != 0xffff);		// skip ports list
				DPMISetSelector(VESA_Selector,*(ulong *)pp,*(ulong *)(pp+2),0xC0F2);
			}
		} else {
			if (set_mode)
				VESA_Display_Win_Func = VESA_Display_Window_16;
		}
	}

	// check to see if linear addressing is supported

	if ((vmi->mode_attr & VMI_MODE_LINEAR_ABLE) && !opt.noVESA_LFB) {
		ulong	siz;

		if (set_mode) {

			siz = (ulong)vmi->bpr * vmi->vres;

			// check to see if we have already allocated a linear buffer and if it is correct

			if (	   VESA_Physical_Window != vmi->phys_baseptr
				|| VESA_Linear_Window_Size < siz) {

				// the window is nonexistent, too small, or in the wrong place

				if (!(VESA_Linear_Window = DPMIPhysMapAttempt(vmi->phys_baseptr,siz)))
					goto use_vga_style_window;

				VESA_Physical_Window = vmi->phys_baseptr;
				VESA_Linear_Window_Size = siz;
			}

			VESA_Use_Linear = TRUE;
		}

	} else {

use_vga_style_window:
	
		VESA_Use_Linear = FALSE;

		// no linear buffer is supported; so there had better be a VGA window

		if (vmi->mode_attr & VMI_MODE_NO_WINDOW)
			return VMCHK_NO_READ_WINDOW;	// aargh

		// mode is supported and a graphics mode... check for a writable window

		if (vmi->win_a_attr & VMI_WIN_SUPPORTED) {
			if (vmi->win_a_attr & VMI_WIN_WRITABLE)
				window=1;	// Use Window A
			if (vmi->win_a_attr & VMI_WIN_READABLE)
				rwindow=1;	// Use Window A
		};

		if (vmi->win_b_attr & VMI_WIN_SUPPORTED) {
			if (!window)
				if (vmi->win_b_attr & VMI_WIN_WRITABLE)
					window=2;	// Use Window B
			if (!rwindow)
				if (vmi->win_b_attr & VMI_WIN_READABLE)
					rwindow=2;	// Use Window B
		};

		if (!rwindow) return VMCHK_NO_READ_WINDOW;	// no readable window
		if (!window) return VMCHK_NO_WRITE_WINDOW;	// no writable window

		if (vmi->win_size<((vmi->bpr+1023)>>10)) return VMCHK_TINY_WINDOW;	// minimum (bpr) window req'd
		if (vmi->win_size & (vmi->win_size-1)) return VMCHK_WEIRD_WINDOW_SIZE;	// Window size must be a power of 2

		if (set_mode) {

			VESA_Window=(uchar *)((window==1 ? vmi->win_a_segment : vmi->win_b_segment)<<4);
			VESA_Window_Size=vmi->win_size<<10;
			VESA_Read_Window=(uchar *)((rwindow==1 ? vmi->win_a_segment : vmi->win_b_segment)<<4);

			VESA_Bank_Shift=0;
			vms=vmi->win_granularity;
			while(vms<vmi->win_size) {
				vms<<=1;
				++VESA_Bank_Shift;
			};

			vms=vmi->win_size;
			VESA_Window_Bits=10;
			while(vms>>=1) ++VESA_Window_Bits;

			VESA_Read_Window_ID=rwindow-1;
			VESA_Window_ID=window-1;
			VESA_Window_Mask=VESA_Window_Size-1;

		}
	}

	if (set_mode)
		VESA_Bpr=vmi->bpr;

	return VMCHK_MODE_FINE;
}

boolean	vidvesa_init(enum VD_InitCode vdic,long x,long y) {
	void *realvm;
	int fault;

	switch(vdic) {
	case VD_STARTUP:
		// get general info about the VESA interface, find out
		// if the controller is VGA compatible

		{
			struct VbeInfoBlock *vinfo;
			ushort desc;
			struct RealInt r;

			if (!(vinfo=(struct VbeInfoBlock *)DosMemoryAllocate(33,&desc)))
			return FALSE;

			memset(vinfo,0,17*16);

			// query VESA

			r.x.eax = 0x4f00;
			r.es = ((ulong)vinfo) >> 4;
			r.w.di = ((ulong)vinfo) & 15;
			DPMIRealInt(0x10,&r);

			if (r.w.ax != 0x004f) return FALSE;

			// check VGA compatibility flag

			if (vinfo->Capabilities & VIB_WEIRDVGA)
				opt.usedirectVGA=FALSE;

			DosMemoryFree(desc);
		}

		// get information about mode

		VESA_Get_Info(VESA_Mode,&vesa_minfo);

		if (fault = VESA_Check_Mode_Acceptable(&vesa_minfo,TRUE))
			*(char **)x = VESA_Mode_Fault_Messages[fault-1];

		vds=&vdvesa_specs;

		if (VESA_Use_Linear) {
			vid_initBitMap(&vidbm,VESA_Linear_Window,0,0,VESA_Width,VESA_Height,VESA_Bpr,piccol);
		} else {
			vidbm.vm=NULL;
			if (!(realvm=malloc(VESA_Width*VESA_Height))) {
				return FALSE;
			}
			vid_initBitMap(&vidbm,(uchar *)realvm,0,0,VESA_Width,VESA_Height,VESA_Width,piccol);
		}
	case VD_REINIT:
	case VD_FROM13:
		regs.x.eax=0x4f02;
		regs.x.ebx=VESA_Mode;
		if (VESA_Use_Linear) regs.x.ebx |= 0x4000;
		int386(0x10,&regs,&regs);

		vid_initBitMap(&realvidbm,(uchar *)0xA0000,0,0,VESA_Width,VESA_Height,VESA_Width,piccol);

		if (VESA_Use_Linear)
			vidbm.syncrect=FALSE;
		else {
			vidbm.syncrect=TRUE;
			vidbm.syncx1=vidbm.syncy1=0x7FFF;
			vidbm.syncx2=vidbm.syncy2=-0x8000;
		}
		maxclr=256;

		break;
	case VD_CLOSEDOWN:
	case VD_TO13:
		vid_setmode(0x13);
		vid_initBitMap(&realvidbm,(uchar *)0xA0000,0,0,320,200,200,piccol);
		vidbm.syncrect=FALSE;
		if (vdic==VD_CLOSEDOWN) {
			if (!VESA_Use_Linear)
				if (vidbm.vm) free(vidbm.vm);
		}
		break;
	case VD_RESIZE:
		return FALSE;
		break;
	case VD_SETSIZE:
		VESA_Mode=x;
		break;
	};
	return TRUE;
}

void	VESA_SetBank(ushort bank);
void	VESA_SetReadBank(ushort bank);

#if 0
#pragma aux VESA_SetBank=		\
	"mov ax,4f05h",			\
	"mov bx,VESA_Window_ID",	\
	"mov cx,VESA_Bank_Shift",	\
	"shl dx,cl",			\
	"int 10h"			\
	parm [dx] modify exact [eax ebx ecx edx];

#pragma aux VESA_SetReadBank=		\
	"mov ax,4f05h",			\
	"mov bx,VESA_Read_Window_ID",	\
	"mov cx,VESA_Bank_Shift",	\
	"shl dx,cl",			\
	"int 10h"			\
	parm [dx] modify exact [eax ebx ecx edx];
#else
#pragma aux VESA_SetBank=		\
	"mov bx,VESA_Window_ID",	\
	"mov cx,VESA_Bank_Shift",	\
	"shl dx,cl",			\
	"push es",			\
	"mov es,VESA_Selector",		\
	"call dword ptr [VESA_Display_Win_Func]",	\
	"pop es"			\
	parm [dx] modify exact [eax ebx ecx edx];
#pragma aux VESA_SetReadBank=		\
	"mov bx,VESA_Read_Window_ID",	\
	"mov cx,VESA_Bank_Shift",	\
	"shl dx,cl",			\
	"push es",			\
	"mov es,VESA_Selector",		\
	"call dword ptr [VESA_Display_Win_Func]",	\
	"pop es"			\
	parm [dx] modify exact [eax ebx ecx edx];
#endif

ushort	VESA_GetBank(void);
#pragma aux VESA_GetBank=		\
	"mov ax,4f05h",			\
	"mov bx,VESA_Window_ID",	\
	"inc bh",			\
	"int 10h",			\
	value [ax] modify exact [eax ebx edx];

ushort	VESA_GetReadBank(void);
#pragma aux VESA_GetReadBank=		\
	"mov ax,4f05h",			\
	"mov bx,VESA_Read_Window_ID",	\
	"inc bh",			\
	"int 10h",			\
	value [ax] modify exact [eax ebx edx];


void	vidvesa_rhline(coord x,coord y,uchar *dm1,coord xs) {
	if (VESA_Use_Linear) {
		qmove(VESA_Linear_Window+x+VESA_Bpr*y,dm1,xs);
	} else {
		uchar *vm1,*vm2,*vm;
		ushort bank;
		ulong vid_add;

		vm1=VESA_Window;
		vm2=VESA_Window+VESA_Window_Size;
		vid_add=VESA_Bpr*y+x;
		bank=vid_add>>VESA_Window_Bits;
		vm=vm1+(vid_add & VESA_Window_Mask);

		VESA_SetBank(bank);

		if ((vm2-vm)>=xs)
			qmove(vm,dm1,xs);
		else {
			dm1=qmove(vm,dm1,vm2-vm);
			VESA_SetBank(++bank);
			dm1=qmove(vm1,dm1,(vid_add+xs) & VESA_Window_Mask);
		}
	}
}

void	vidvesa_rhlineread(coord x,coord y,uchar *dm1,coord xs) {
	if (VESA_Use_Linear) {
		qmove(dm1,VESA_Linear_Window+x+VESA_Bpr*y,xs);
	} else {
		uchar *vm1,*vm2,*vm;
		ushort bank;
		ulong vid_add;
		ulong s;

		vm1=VESA_Read_Window;
		vm2=VESA_Read_Window+VESA_Window_Size;
		vid_add=VESA_Bpr*y+x;
		bank=vid_add>>VESA_Window_Bits;
		vm=vm1+(vid_add & VESA_Window_Mask);

		VESA_SetReadBank(bank);

		if ((vm2-vm)>=xs)
			qmove(vm,dm1,xs);
		else {
			qmove(dm1,vm,s=(vm2-vm));
			dm1+=s;
			VESA_SetReadBank(++bank);
			dm1=qmove(dm1,vm1,(vid_add+xs) & VESA_Window_Mask);
		}
	}
}

void	vidvesa_rbitblt(coord x,coord y,coord xs,coord ys,uchar *dm1,ulong dmod,boolean no_offset) {
	if (VESA_Use_Linear) {
		ulong offs = x+y*VESA_Bpr;

		if (!no_offset) dm1+=x+y*VESA_Width;

		rectcopy(VESA_Linear_Window+offs,VESA_Bpr,dm1,dmod,xs,ys);
	} else {
		uchar *vm1,*vm2,*vm;
		ushort bank;
		ulong vid_add;
		short dbpr;

		if (xs<=0) return;
		if (ys<=0) return;

		if (!no_offset) dm1+=vidbm.bpr*y+x;

		vm1=VESA_Window;
		vm2=VESA_Window+VESA_Window_Size;
		vid_add=VESA_Bpr*y+x;
		bank=vid_add>>VESA_Window_Bits;
		vm=vm1+(vid_add & VESA_Window_Mask);
		dbpr=dmod-xs;

		VESA_SetBank(bank);

		do {
			if ((vm2-vm)>=xs) {
				dm1=qmove(vm,dm1,xs);
				if ((vm+=VESA_Bpr) >= vm2) {
					VESA_SetBank(++bank);
					vm-=VESA_Window_Size;
				};
			} else {
				dm1=qmove(vm,dm1,vm2-vm);
				VESA_SetBank(++bank);
				dm1=qmove(vm1,dm1,vm+xs-vm2);
				vm+=VESA_Bpr;
				vm-=VESA_Window_Size;
			};
			dm1+=dbpr;
		} while(--ys);
	}

//	do {
//		vidvesa_rhline(x,y++,dm1,xs);
//		dm1+=vidbm.bpr;
//	} while(--ys);
}

void	vidvesa_rbitbltread(uchar *dm1,coord x,coord y,coord xs,coord ys,long dmod,boolean no_offset) {
	if (VESA_Use_Linear) {
		ulong offs = x+y*VESA_Bpr;

		if (!no_offset) dm1+=x+y*VESA_Width;

		rectcopy(dm1,dmod,VESA_Linear_Window+offs,VESA_Bpr,xs,ys);
	} else {
		uchar *vm1,*vm2,*vm;
		ushort bank;
		ulong vid_add;
		long dbpr;

		if (xs<=0) return;
		if (ys<=0) return;

		if (!no_offset) dm1+=vidbm.bpr*y+x;

		vm1=VESA_Read_Window;
		vm2=VESA_Read_Window+VESA_Window_Size;
		vid_add=VESA_Bpr*y+x;
		bank=vid_add>>VESA_Window_Bits;
		vm=vm1+(vid_add & VESA_Window_Mask);

		dbpr=dmod;

		VESA_SetReadBank(bank);

		do {
			if ((vm2-vm)>=xs) {
				qmove(dm1,vm,xs);
				if ((vm+=VESA_Bpr) >= vm2) {
					VESA_SetReadBank(++bank);
					vm-=VESA_Window_Size;
				};
			} else {
				qmove(dm1,vm,vm2-vm);
				VESA_SetReadBank(++bank);
				qmove(dm1+(vm2-vm),vm1,(vm+xs-vm2) & VESA_Window_Mask);
				vm+=VESA_Bpr;
				vm-=VESA_Window_Size;
			};
			dm1+=dbpr;
		} while(--ys);
	}
}

/* be very careful with the pointer functions -- they are called
** when SS != DS, and ES != DS.  That means use qmovebes() and no
** pointers to automatic variables. */

#pragma off (check_stack)

void	vidvesa_drawptr(struct MousePointerInfo *mpi,coord x,coord y) {
	ushort xs=mpi->s_width,ys=mpi->s_height,xst,yst;
	uchar *dp1,*dp2,c;
	ushort readbank,writebank;

	dp1=mpi->image;

	if (x<=-mpi->s_width) goto abort;
	else if (x>=vidbm.width) goto abort;
	else if (x>(vidbm.width-mpi->s_width)) xs=vidbm.width-x;
	else if (x<0) { xs=mpi->s_width+x; dp1-=x; x=0; }

	if (y<=-mpi->s_height) goto abort;
	else if (y>=vidbm.height) goto abort;
	else if (y>(vidbm.height-mpi->s_height)) ys=vidbm.height-y;
	else if (y<0) { ys=mpi->s_height+y; dp1-=y*mpi->width; y=0; }

	mpi->width	= xs;
	mpi->height	= ys;
	mpi->saved_x	= x;
	mpi->saved_y	= y;

	if (xs<=0 || ys<=0) goto abort;

	if (!VESA_Use_Linear) {
		if (VESA_Window_ID != VESA_Read_Window_ID) readbank=VESA_GetReadBank();
		writebank=VESA_GetBank();
	}

	vidvesa_rbitbltread(mpi->backup,x,y,xs,ys,xs,TRUE);

	qmovel((void*)mpi->blitmem,(void*)mpi->backup,(mpi->s_width*mpi->s_height)/4);

	dp2=mpi->blitmem;

	yst=ys;
	do {
		xst=xs;
		do {
			if (c=*dp1) *dp2=c-1 ? mpi->high : mpi->low;
			++dp1,++dp2;
		} while(--xst);
		dp1+=(mpi->s_width-xs);
	} while(--yst);

	vidvesa_rbitblt(x,y,xs,ys,mpi->blitmem,xs,TRUE);

	if (!VESA_Use_Linear) {
		if (VESA_Window_ID != VESA_Read_Window_ID) VESA_SetReadBank(readbank);
		VESA_SetBank(writebank);
	}

	return;

	// avoids a function :-)

abort:
	mpi->width=mpi->height=0;
}

void	vidvesa_eraseptr(struct MousePointerInfo *mpi) {
	ushort xs,ys;
	ushort bank;

	if ((xs=mpi->width)<=0) return;
	if ((ys=mpi->height)<=0) return;

	if (!VESA_Use_Linear)	bank=VESA_GetBank();

	vidvesa_rbitblt(mpi->saved_x,mpi->saved_y,xs,ys,mpi->backup,xs,TRUE);

	if (!VESA_Use_Linear)	VESA_SetBank(bank);
}


